import copy
import os
import numpy as np
import pandas as pd

from .plotting import plot_cats, plot_bounds


class SelectionRulePicker:
    def __init__(self, distance_from_best_table):
        """Runs simulations of different model group selection rules

        Can look at different results of selection rules, like 'regrets'
        or raw metric values.

        A regret refers to the difference in performance between a model group
        and the best model group for the next testing window
        if a selection rule is followed.

        Args:
            distance_from_best_table (audition.DistanceFromBestTable)
                A pre-populated distance-from-best database table
        """
        self.distance_from_best_table = distance_from_best_table

    def results_for_rule(
        self,
        bound_selection_rule,
        model_group_ids,
        train_end_times,
        regret_metric,
        regret_parameter,
    ):
        """Calculate the regrets, or distance between the chosen model and
            the maximum value next test time

        Arguments:
            bound_selection_rule (function) A function that returns a model group
                given a dataframe of model group performances plus other
                arguments
            model_group_ids (list) The list of model group ids to include in
                the regret analysis
            train_end_times (list) The list of train end times to include in
                the regret analysis
            regret_metric (string) -- model evaluation metric, such as 'precision@'
            regret_parameter (string) -- model evaluation metric parameter,
                such as '300_abs'

        Returns: (list) for each train end time, a dictionary representing the
            model group chosen by the selection rule and its performance

            Should have all keys from the DistanceFromBestTable, most
            importantly
            'distance_from_best_case_next_time'
            'distance_from_best_case',
            'raw_value',
            'raw_value_next_time'
        """
        df = self.distance_from_best_table.as_dataframe(model_group_ids)
        choices = []
        for train_end_time in train_end_times:
            # When plotting rules, we use only the best one model group id
            model_group_id = self.model_group_from_rule(
                bound_selection_rule, model_group_ids, train_end_time
            )
            model_group_id = model_group_id[0]
            choice = df[
                (df["model_group_id"] == model_group_id)
                & (df["train_end_time"] == train_end_time)
                & (df["metric"] == regret_metric)
                & (df["parameter"] == regret_parameter)
            ]
            assert len(choice) == 1
            choices.append(choice.squeeze().to_dict())
        return choices

    def model_group_from_rule(
        self, bound_selection_rule, model_group_ids, train_end_time
    ):
        """Pick a model group that best selects the given selection rule

        In here, we create a subset of the distance from best table dataframe,
        with all data after the given train end time removed, both rows representing
        later time periods but also columns that have access to later data. Calculating and
        passing this allows the selection rules to be written without specific code
        to exclude the future

        Arguments:
            bound_selection_rule (function) A function that returns a model group
                given a dataframe of model group performances plus other
                arguments
            model_group_ids (list) The list of model group ids to consider
            train_end_time (timestamp) The list of train end times to include in
                the regret analysis

        Returns: (int) The model group id chosen by the input selection rule
        """
        df = self.distance_from_best_table.as_dataframe(model_group_ids)
        localized_df = copy.deepcopy(df[df["train_end_time"] <= train_end_time])
        del localized_df["dist_from_best_case_next_time"]

        return bound_selection_rule.pick(localized_df, train_end_time)


class SelectionRulePlotter:
    """Plot selection rules

    Args:
        selection_rule_picker (.SelectionRulePicker)
    """

    def __init__(self, selection_rule_picker, directory=None):
        self.selection_rule_picker = selection_rule_picker
        self.directory = directory

    def plot_bounds(self, metric, parameter):
        """Compute the plot bounds for a given metric and parameter

        Args:
            metric (string) -- model evaluation metric, such as 'precision@'
            parameter (string) -- model evaluation metric parameter,
                such as '300_abs'

        Returns: (number, number) A minimum and maximum x value for the plot
        """

        observed_min, observed_max = \
            self.selection_rule_picker.distance_from_best_table.observed_bounds[(metric, parameter)]
        return plot_bounds(observed_min, observed_max)

    def regret_threshold_dist(self, plot_min, plot_max):
        dist = plot_max - plot_min
        return dist / 100.0

    def regret_thresholds(self, regret_metric, regret_parameter):
        plot_min, plot_max = self.plot_bounds(regret_metric, regret_parameter)
        regret_threshold_dist = self.regret_threshold_dist(plot_min, plot_max)
        return np.arange(plot_min, plot_max, regret_threshold_dist)

    def create_plot_dataframe(
        self,
        bound_selection_rules,
        model_group_ids,
        train_end_times,
        regret_metric,
        regret_parameter,
    ):
        """Create a dataframe suitable for plotting selection rule regrets

        Args:
            bound_selection_rules (list of .selection_rules.BoundSelectionRule)
               Selection rule/parameter combinations to plot
            model_group_ids (list of integers) The model group ids to include
                in calculating the data
            train_end_times (list of timestamps) The timestamps to include in
                calculating the data
            regret_metric (string) The metric (i.e. precision@) to calculate
                regrets against
            regret_parameter (string) The metric parameter (i.e. 100_abs) to
                calculate regrets against

        Returns: (pandas.DataFrame) A dataframe with columns 'regret',
            'pct_of_time', and 'selection_rule'
        """
        accumulator = []
        for selection_rule in bound_selection_rules:
            regrets = [
                result["dist_from_best_case_next_time"]
                for result in self.selection_rule_picker.results_for_rule(
                    selection_rule,
                    model_group_ids,
                    train_end_times,
                    regret_metric,
                    regret_parameter,
                )
            ]
            for regret_threshold in self.regret_thresholds(
                regret_metric, regret_parameter
            ):
                distro = [int(regret < regret_threshold) for regret in regrets]
                accumulator.append(
                    {
                        "regret": regret_threshold,
                        "pct_of_time": np.mean(distro),
                        "selection_rule": selection_rule.descriptive_name,
                    }
                )
        return pd.DataFrame.from_records(accumulator)

    def plot_all_selection_rules(
        self,
        bound_selection_rules,
        model_group_ids,
        train_end_times,
        regret_metric,
        regret_parameter,
    ):
        """Plot the regrets of all given selection rules

        Args:
            bound_selection_rules (list of .selection_rules.BoundSelectionRule)
               Selection rule/parameter combinations to plot
            model_group_ids (list of integers) The model group ids to include
                in calculating the data
            train_end_times (list of timestamps) The timestamps to include in
                calculating the data
            regret_metric (string) The metric (i.e. precision@) to calculate
                regrets against
            regret_parameter (string) The metric parameter (i.e. 100_abs) to
                calculate regrets against
        """
        df_regrets = self.create_plot_dataframe(
            bound_selection_rules,
            model_group_ids,
            train_end_times,
            regret_metric,
            regret_parameter,
        )
        cat_col = "selection_rule"
        plt_title = "Fraction of models X pp worse than best {} {} next time".format(
            regret_metric, regret_parameter
        )
        (plot_min, plot_max) = self.plot_bounds(regret_metric, regret_parameter)

        if self.directory:
            path_to_save = os.path.join(
                self.directory,
                f"regret_distance_from_best_rules_{regret_metric}{regret_parameter}.png",
            )
        else:
            path_to_save = None

        plot_cats(
            frame=df_regrets,
            x_col="regret",
            y_col="pct_of_time",
            cat_col=cat_col,
            grp_col="selection_rule",
            title=plt_title,
            x_label="distance from best {} next time".format(regret_metric),
            y_label="fraction of models",
            x_lim=self.plot_bounds(regret_metric, regret_parameter),
            path_to_save=path_to_save,
            alpha=1.0,
        )



'''
A4uxiliary functions and helpers:

This set of functions are helper functions to format data
(i.e., prediction matrices, etc.) for plotting. This functions
are called in both Model class and ModelGroup class in
evaluation.py.
'''

from sqlalchemy import create_engine
from sqlalchemy.sql import text
from collections import namedtuple
import yaml

import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)


ModelEvaluator = namedtuple('ModelEvaluator',
                           ('model_group_id', 'model_id'))


def create_pgconn(credentials_yaml):
    '''
    Create SQL connection object using a psycopg2 cursor and abiding to new
    dssg/dsapp db user configuration.

    Arguments:
        - credentials_yaml: .yaml file with db credentials
    '''
    with open(credentials_yaml) as f:
        configs = yaml.full_load(f)
    try:
        conn = create_engine("postgresql://{user}:{password}@{host}:{port}/{dbname}".format(**configs))
    except:
        logger.error("Error connecting to db.")

    return conn


def get_models_ids(audited_model_group_ids, conn):
    '''
    This helper functions will retrieve the model_id's from a set
    of model_group_ids and will instantiate each model into the
    ModelEvaluator class.

    Aguments:
        - audited_model_group_ids: List of model_group_ids
          (ideally from Audition's output)
        - conn: sql engine

    This function will return a list of ModelEvaluator objects
    '''

    query = conn.execute(text("""
    SELECT model_group_id,
           model_id
    FROM triage_metadata.models
    WHERE model_group_id = ANY(:ids);
    """), ids=audited_model_group_ids)

    return [ModelEvaluator._make(row) for row in query]



from sqlalchemy import Table
from sqlalchemy import create_engine
from sqlalchemy.types import VARCHAR
from testing.postgresql import Postgresql
from unittest import TestCase

from triage.component.architect import database_reflection as dbreflect


class TestDatabaseReflection(TestCase):
    def setUp(self):
        self.postgresql = Postgresql()
        self.engine = create_engine(self.postgresql.url())

    def tearDown(self):
        self.postgresql.stop()

    def test_split_table(self):
        assert dbreflect.split_table("staging.incidents") == ("staging", "incidents")
        assert dbreflect.split_table("incidents") == (None, "incidents")
        with self.assertRaises(ValueError):
            dbreflect.split_table("blah.staging.incidents")

    def test_table_object(self):
        assert isinstance(dbreflect.table_object("incidents", self.engine), Table)

    def test_reflected_table(self):
        self.engine.execute("create table incidents (col1 varchar)")
        assert dbreflect.reflected_table("incidents", self.engine).exists()

    def test_table_exists(self):
        self.engine.execute("create table incidents (col1 varchar)")
        assert dbreflect.table_exists("incidents", self.engine)
        assert not dbreflect.table_exists("compliments", self.engine)

    def test_table_has_data(self):
        self.engine.execute("create table incidents (col1 varchar)")
        self.engine.execute("create table compliments (col1 varchar)")
        self.engine.execute("insert into compliments values ('good job')")
        assert dbreflect.table_has_data("compliments", self.engine)
        assert not dbreflect.table_has_data("incidents", self.engine)

    def test_table_has_column(self):
        self.engine.execute("create table incidents (col1 varchar)")
        assert dbreflect.table_has_column("incidents", "col1", self.engine)
        assert not dbreflect.table_has_column("incidents", "col2", self.engine)

    def test_column_type(self):
        self.engine.execute("create table incidents (col1 varchar)")
        assert dbreflect.column_type("incidents", "col1", self.engine) == VARCHAR



from triage.component.architect.features import FeatureGroupCreator
from unittest import TestCase


def test_table_group():
    group = FeatureGroupCreator(definition={"tables": ["one", "three"]})

    subsets = group.subsets(
        {
            "one": ["col_a", "col_b", "col_c"],
            "two": ["col_a", "col_b", "col_c"],
            "three": ["col_a", "col_b", "col_c"],
        }
    )
    assert subsets == [
        {"one": ["col_a", "col_b", "col_c"]},
        {"three": ["col_a", "col_b", "col_c"]},
    ]
    assert subsets[0].names == ["tables: one"]
    assert subsets[1].names == ["tables: three"]


def test_prefix_group():
    # ensure we test prefixes with underscores
    group = FeatureGroupCreator(definition={"prefix": ["major_viol", "severe_viol"]})

    subsets = group.subsets(
        {
            "one": ["minor_viol_a", "minor_viol_b", "minor_viol_c"],
            "two": ["severe_viol_a", "severe_viol_b", "severe_viol_c"],
            "three": ["major_viol_a", "major_viol_b", "major_viol_c"],
            "four": ["minor_viol_a", "minor_viol_b", "minor_viol_c"],
        }
    )
    assert subsets == [
        {"three": ["major_viol_a", "major_viol_b", "major_viol_c"]},
        {"two": ["severe_viol_a", "severe_viol_b", "severe_viol_c"]},
    ]
    assert subsets[0].names == ["prefix: major_viol"]
    assert subsets[1].names == ["prefix: severe_viol"]


def test_multiple_criteria():
    group = FeatureGroupCreator(
        definition={"prefix": ["major", "severe"], "tables": ["one", "two"]}
    )

    subsets = group.subsets(
        {
            "one": ["minor_a", "minor_b", "minor_c"],
            "two": ["severe_a", "severe_b", "severe_c"],
            "three": ["major_a", "major_b", "major_c"],
            "four": ["minor_a", "minor_b", "minor_c"],
        }
    )
    assert subsets == [
        {"three": ["major_a", "major_b", "major_c"]},
        {"two": ["severe_a", "severe_b", "severe_c"]},
        {"one": ["minor_a", "minor_b", "minor_c"]},
        {"two": ["severe_a", "severe_b", "severe_c"]},
    ]

    assert subsets[0].names == ["prefix: major"]
    assert subsets[1].names == ["prefix: severe"]
    assert subsets[2].names == ["tables: one"]
    assert subsets[3].names == ["tables: two"]


class TestValidations(TestCase):
    def test_not_a_dict(self):
        creator = FeatureGroupCreator(definition="prefix")
        with self.assertRaises(ValueError):
            creator.validate()

    def test_badsubsetter(self):
        creator = FeatureGroupCreator(
            definition={"prefix": ["major", "severe"], "other": ["other"]}
        )
        with self.assertRaises(ValueError):
            creator.validate()

    def test_stringvalue(self):
        creator = FeatureGroupCreator(definition={"prefix": "major"})
        with self.assertRaises(ValueError):
            creator.validate()

    def test_goodconfig(self):
        creator = FeatureGroupCreator(
            definition={"prefix": ["major", "severe"], "tables": ["one", "two"]}
        )
        creator.validate()



import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

from numbers import Number
from itertools import product, chain
import sqlalchemy.sql.expression as ex
import re
from descriptors import cachedproperty

from .sql import make_sql_clause, to_sql_name, CreateTableAs, InsertFromSelect
from .imputations import (
    ImputeMean,
    ImputeConstant,
    ImputeZero,
    ImputeZeroNoFlag,
    ImputeNullCategory,
    ImputeBinaryMode,
    ImputeError,
)

available_imputations = {
    "mean": ImputeMean,
    "constant": ImputeConstant,
    "zero": ImputeZero,
    "zero_noflag": ImputeZeroNoFlag,
    "null_category": ImputeNullCategory,
    "binary_mode": ImputeBinaryMode,
    "error": ImputeError,
}


def make_list(a):
    return [a] if not isinstance(a, list) else a


def make_tuple(a):
    return (a,) if not isinstance(a, tuple) else a


DISTINCT_REGEX = re.compile(r"distinct[ (]")
AGGFUNCS_NEED_MULTIPLE_VALUES = set(['stddev', 'stddev_samp', 'variance', 'var_samp'])


def split_distinct(quantity):
    # Only support distinct clauses with one-argument quantities
    if len(quantity) != 1:
        return ("", quantity)
    q = quantity[0]
    if DISTINCT_REGEX.match(q):
        return "distinct ", (q[8:].lstrip(" "),)
    else:
        return "", (q,)


class AggregateExpression:
    def __init__(
        self,
        aggregate1,
        aggregate2,
        operator,
        cast=None,
        operator_str=None,
        expression_template=None,
    ):
        """
        Args:
            aggregate1: first aggregate
            aggregate2: second aggregate
            operator: string of SQL operator, e.g. "+"
            cast: optional string to put after aggregate1, e.g. "*1.0", "::decimal"
            operator_str: optional name of operator to use, defaults to operator
            expression_template: optional formatting template with the following keywords:
                name1, operator, name2
        """
        self.aggregate1 = aggregate1
        self.aggregate2 = aggregate2
        self.operator = operator
        self.cast = cast if cast else ""
        self.operator_str = operator if operator_str else operator
        self.expression_template = (
            expression_template if expression_template else "{name1}{operator}{name2}"
        )

    def column_imputation_lookup(self, prefix=None):
        lookup1 = self.aggregate1.column_imputation_lookup()
        lookup2 = self.aggregate2.column_imputation_lookup()

        return dict(
            (
                prefix
                + self.expression_template.format(
                    name1=c1, operator=self.operator_str, name2=c2
                ),
                lookup1[c1],
            )
            for c1, c2 in product(lookup1.keys(), lookup2.keys())
        )

    def alias(self, expression_template):
        """
        Set the expression template used for naming columns of an AggregateExpression
        Returns: self, for chaining
        """
        self.expression_template = expression_template
        return self

    def get_columns(self, when=None, prefix=None, format_kwargs=None):
        if prefix is None:
            prefix = ""

        columns1 = self.aggregate1.get_columns(when)
        columns2 = self.aggregate2.get_columns(when)

        for c1, c2 in product(columns1, columns2):
            c = ex.literal_column(
                "({}{} {} {})".format(c1, self.cast, self.operator, c2)
            )
            yield c.label(
                prefix
                + self.expression_template.format(
                    name1=c1.name, operator=self.operator_str, name2=c2.name
                )
            )

    def __add__(self, other):
        return AggregateExpression(self, other, "+")

    def __sub__(self, other):
        return AggregateExpression(self, other, "-")

    def __mul__(self, other):
        return AggregateExpression(self, other, "*")

    def __div__(self, other):
        return AggregateExpression(self, other, "/", "*1.0")

    def __truediv__(self, other):
        return AggregateExpression(self, other, "/", "*1.0")

    def __lt__(self, other):
        return AggregateExpression(self, other, "<")

    def __le__(self, other):
        return AggregateExpression(self, other, "<=")

    def __eq__(self, other):
        return AggregateExpression(self, other, "=")

    def __ne__(self, other):
        return AggregateExpression(self, other, "!=")

    def __gt__(self, other):
        return AggregateExpression(self, other, ">")

    def __ge__(self, other):
        return AggregateExpression(self, other, ">=")

    def __or__(self, other):
        return AggregateExpression(self, other, "or", operator_str="|")

    def __and__(self, other):
        return AggregateExpression(self, other, "and", operator_str="&")


class Aggregate(AggregateExpression):
    """
    An object representing one or more SQL aggregate columns in a groupby
    """

    def __init__(self, quantity, function, impute_rules, order=None, coltype=None):
        """
        Args:
            quantity: SQL for the quantity to aggregate
            function: SQL aggregate function
            impute_rules: dictionary of rules mapping functions to imputation methods
            order: SQL for order by clause in an ordered set aggregate
            coltype: SQL type for the column in the generated features table

        Notes:
            quantity, function, and order can also be lists of the above,
            in which case the cross product of those is used. If quantity is a
            collection than name should also be a collection of the same length.

            quantity can be a tuple of SQL quantities for aggregate functions
            that take multiple arguments, e.g. corr, regr_slope

            quantity can be a dictionary in which case the keys are names
            for the expressions and values are expressions.
        """
        if isinstance(quantity, dict):
            # make quantity values tuples
            self.quantities = {k: make_tuple(q) for k, q in quantity.items()}
        else:
            # first convert to list of tuples
            quantities = [make_tuple(q) for q in make_list(quantity)]
            # then dict with name keys
            self.quantities = {to_sql_name(str.join("_", q)): q for q in quantities}

        self.functions = make_list(function)
        self.orders = make_list(order)
        self.impute_rules = impute_rules
        self.coltype = coltype

    def get_columns(self, when=None, prefix=None, format_kwargs=None):
        """
        Args:
            when: used in a case statement to filter the rows going into the
                aggregation function
            prefix: prefix for column names
            format_kwargs: kwargs to pass to format the aggregate quantity
        Returns:
            collection of SQLAlchemy columns
        """
        if prefix is None:
            prefix = ""
        if format_kwargs is None:
            format_kwargs = {}

        name_template = "{prefix}{quantity_name}_{function}"
        coltype_template = ""
        column_template = "{function}({distinct}{args}){order_clause}{filter}{coltype_cast}"
        arg_template = "{quantity}"
        order_template = ""
        filter_template = ""

        if self.orders != [None]:
            order_template += " WITHIN GROUP (ORDER BY {order})"
        if when:
            filter_template = " FILTER (WHERE {when})"

        if self.coltype is not None:
            coltype_template = "::{coltype}"

        for function, (quantity_name, quantity), order in product(
            self.functions, self.quantities.items(), self.orders
        ):
            distinct, quantity = split_distinct(quantity)
            args = str.join(", ", (arg_template.format(quantity=q) for q in quantity))
            order_clause = order_template.format(order=order)
            filter = filter_template.format(when=when)
            coltype_cast = coltype_template.format(coltype=self.coltype)

            if order is not None:
                if len(quantity_name) > 0:
                    quantity_name += "_"
                quantity_name += to_sql_name(order)

            kwargs = dict(
                function=function,
                args=args,
                prefix=prefix,
                distinct=distinct,
                order_clause=order_clause,
                quantity_name=quantity_name,
                filter=filter,
                coltype_cast=coltype_cast,
                **format_kwargs
            )

            column = column_template.format(**kwargs).format(**format_kwargs)
            name = name_template.format(**kwargs)

            yield ex.literal_column(column).label(to_sql_name(name))

    def column_imputation_lookup(self, prefix=None):
        """
        Args:
            prefix: prefix for column names
        Returns:
            dictionary mapping columns to appropriate imputation rule
        """
        if prefix is None:
            prefix = ""

        name_template = "{prefix}{quantity_name}_{function}"

        lkup = {}
        for function, (quantity_name, quantity), order in product(
            self.functions, self.quantities.items(), self.orders
        ):

            if order is not None:
                if len(quantity_name) > 0:
                    quantity_name += "_"
                quantity_name += to_sql_name(order)

            kwargs = dict(function=function, prefix=prefix, quantity_name=quantity_name)

            name = name_template.format(**kwargs)

            # requires an imputation rule defined for any function
            # type used by the aggregate (or catch-all with 'all')
            try:
                lkup[name] = dict(
                    self.impute_rules.get(function) or self.impute_rules["all"],
                    coltype=self.impute_rules["coltype"],
                )
            except KeyError as err:
                raise ValueError(
                    "Must provide an imputation rule for every aggregation "
                    + "function (or 'all'). No rule found for %s" % name
                ) from err

        return lkup


def maybequote(elt, quote_override=None):
    "Quote for passing to SQL if necessary, based upon the python type"

    def quote_string(string):
        return "'{}'".format(string)

    if quote_override is None:
        if isinstance(elt, Number):
            return elt
        else:
            return quote_string(elt)
    elif quote_override:
        return quote_string(elt)
    else:
        return elt


class Compare(Aggregate):
    """
    A simple shorthand to automatically create many comparisons against one column
    """

    def __init__(
        self,
        col,
        op,
        choices,
        function,
        impute_rules,
        order=None,
        coltype=None,
        include_null=False,
        maxlen=None,
        op_in_name=True,
        quote_choices=None,
    ):
        """
        Args:
            col: the column name (or equivalent SQL expression)
            op: the SQL operation (e.g., '=' or '~' or 'LIKE')
            choices: A list or dictionary of values. When a dictionary is
                passed, the keys are a short name for the value.
            function: (from Aggregate)
            impute_rules: (from Aggregate)
            order: (from Aggregate)
            include_null: Add an extra `{col} is NULL` if True (default False).
                 May also be non-boolean, in which case its truthiness determines
                 the behavior and the value is used as the value short name.
            maxlen: The maximum length of aggregate quantity names, if specified.
                Names longer than this will be truncated.
            op_in_name: Include the operator in aggregate names (default False)
            quote_choices: Override smart quoting if present (default None)

        A simple helper method to easily create many comparison columns from
        one source column by comparing it against many values. It effectively
        creates many quantities of the form "({col} {op} {elt})::INT" for elt
        in choices. It automatically quotes strings appropriately and leaves
        numbers unquoted. The type of the comparison is converted to an
        integer so it can easily be used with 'sum' (for total count) and
        'avg' (for relative fraction) aggregate functions.

        By default, the aggregates are named "{col}_{op}_{elt}", but the
        operator may be ommitted if `op_in_name=False`. This name can become
        long and exceed the maximum column name length. If ``maxlen`` is
        specified then any aggregate name longer than ``maxlen`` gets
        truncated with a number appended to ensure that they remain unique and
        identifiable (but note that sequntial ordering is not preserved).
        """
        if type(choices) is not dict:
            choices = {k: k for k in choices}
        opname = "_{}_".format(op) if op_in_name else "_"
        d = {
            "{}{}{}".format(col, opname, nickname): "({} {} {})::INT".format(
                col, op, maybequote(choice, quote_choices)
            )
            for nickname, choice in choices.items()
        }
        if include_null is True:
            include_null = "_NULL"
        if include_null:
            d["{}_{}".format(col, include_null)] = "({} is NULL)::INT".format(col)
        if maxlen is not None and any(len(k) > maxlen for k in d.keys()):
            for i, k in enumerate(list(d.keys())):
                d["%s_%02d" % (k[: maxlen - 3], i)] = d.pop(k)

        Aggregate.__init__(self, d, function, impute_rules, order, coltype)


class Categorical(Compare):
    """
    A simple shorthand to automatically create many equality comparisons against one column
    """

    def __init__(
        self,
        col,
        choices,
        function,
        impute_rules,
        order=None,
        op_in_name=False,
        coltype=None,
        **kwargs
    ):
        """
        Create a Compare object with an equality operator, ommitting the `=`
        from the generated aggregation names. See Compare for more details.

        As a special extension, Compare's 'include_null' keyword option may be
        enabled by including the value `None` in the choices list. Multiple
        None values are ignored.
        """
        if None in choices:
            kwargs["include_null"] = True
            choices.remove(None)
        elif type(choices) is dict and None in choices.values():
            ks = [k for k, v in choices.items() if v is None]
            for k in ks:
                choices.pop(k)
                kwargs["include_null"] = str(k)
        Compare.__init__(
            self,
            col,
            "=",
            choices,
            function,
            impute_rules,
            order,
            coltype,
            op_in_name=op_in_name,
            **kwargs
        )


class Aggregation:
    def __init__(
        self,
        aggregates,
        groups,
        from_obj,
        state_table,
        state_group=None,
        prefix=None,
        suffix=None,
        schema=None,
    ):
        """
        Args:
            aggregates: collection of Aggregate objects.
            from_obj: defines the from clause, e.g. the name of the table. can use
            groups: a list of expressions to group by in the aggregation or a dictionary
                pairs group: expr pairs where group is the alias (used in column names)
            state_table: schema.table to query for comprehensive set of state_group entities
                regardless of what exists in the from_obj
            state_group: the group level found in the state table (e.g., "entity_id")
            prefix: prefix for aggregation tables and column names, defaults to from_obj
            suffix: suffix for aggregation table, defaults to "aggregation"
            schema: schema for aggregation tables

        The from_obj and group expressions are passed directly to the
            SQLAlchemy Select object so could be anything supported there.
            For details see:
            http://docs.sqlalchemy.org/en/latest/core/selectable.html

        Aggregates will have {collate_date} in their quantities substituted with the date
        of aggregation.
        """
        self.aggregates = aggregates
        self.from_obj = make_sql_clause(from_obj, ex.text)
        self.groups = (
            groups if isinstance(groups, dict) else {str(g): g for g in groups}
        )
        self.state_table = state_table
        self.state_group = state_group if state_group else "entity_id"
        self.prefix = prefix if prefix else str(from_obj)
        self.suffix = suffix if suffix else "aggregation"
        self.schema = schema

    @cachedproperty
    def colname_aggregate_lookup(self):
        """A reverse lookup from column name to the source collate.Aggregate

        Will error if the Aggregation contains duplicate column names
        """
        lookup = {}
        for group, groupby in self.groups.items():
            for agg in self.aggregates:
                for col in agg.get_columns(prefix=self._col_prefix(group)):
                    if col.name in lookup:
                        raise ValueError("Duplicate feature column name found: ", col.name)
                    lookup[col.name] = agg
        return lookup

    def _col_prefix(self, group):
        """
        Helper for creating a column prefix for the group
            group: group clause, for naming columns
        Returns: string for a common column prefix for columns in that group
        """
        return "{prefix}_{group}_".format(prefix=self.prefix, group=group)

    def _get_aggregates_sql(self, group):
        """
        Helper for getting aggregates sql
        Args:
            group: group clause, for naming columns
        Returns: collection of aggregate column SQL strings
        """
        return chain(*[a.get_columns(prefix=self._col_prefix(group)) for a in self.aggregates])

    def get_selects(self):
        """
        Constructs select queries for this aggregation

        Returns: a dictionary of group : queries pairs where
            group are the same keys as groups
            queries is a list of Select queries, one for each date in dates
        """
        queries = {}

        for group, groupby in self.groups.items():
            columns = [make_sql_clause(groupby, ex.text)]
            columns += self._get_aggregates_sql(group)

            gb_clause = make_sql_clause(groupby, ex.literal_column)
            query = ex.select(columns=columns, from_obj=make_sql_clause(self.from_obj, ex.text)).group_by(
                gb_clause
            )

            queries[group] = [query]

        return queries

    def get_imputation_rules(self):
        """
        Constructs a dictionary to lookup an imputation rule from an associated
        column name.

        Returns: a dictionary of column : imputation_rule pairs
        """
        imprules = {}
        for group, groupby in self.groups.items():
            prefix = "{prefix}_{group}_".format(prefix=self.prefix, group=group)
            for a in self.aggregates:
                imprules.update(a.column_imputation_lookup(prefix=prefix))
        return imprules

    def get_table_name(self, group=None, imputed=False):
        """
        Returns name for table for the given group
        """
        if group is None and not imputed:
            name = '"%s_%s"' % (self.prefix, self.suffix)
        elif group is None and imputed:
            name = '"%s_%s_%s"' % (self.prefix, self.suffix, "imputed")
        elif imputed:
            name = '"%s"' % to_sql_name("%s_%s_%s" % (self.prefix, group, "imputed"))
        else:
            name = '"%s"' % to_sql_name("%s_%s" % (self.prefix, group))
        schema = '"%s".' % self.schema if self.schema else ""
        return "%s%s" % (schema, name)

    def get_creates(self):
        """
        Construct create queries for this aggregation
        Args:
            selects: the dictionary of select queries to use
                if None, use self.get_selects()
                this allows you to customize select queries before creation

        Returns:
            a dictionary of group : create pairs where
                group are the same keys as groups
                create is a CreateTableAs object
        """
        return {
            group: CreateTableAs(self.get_table_name(group), next(iter(sels)).limit(0))
            for group, sels in self.get_selects().items()
        }

    def get_inserts(self):
        """
        Construct insert queries from this aggregation
        Args:
            selects: the dictionary of select queries to use
                if None, use self.get_selects()
                this allows you to customize select queries before creation

        Returns:
            a dictionary of group : inserts pairs where
                group are the same keys as groups
                inserts is a list of InsertFromSelect objects
        """
        return {
            group: [InsertFromSelect(self.get_table_name(group), sel) for sel in sels]
            for group, sels in self.get_selects().items()
        }

    def get_drops(self):
        """
        Generate drop queries for this aggregation

        Returns: a dictionary of group : drop pairs where
            group are the same keys as groups
            drop is a raw drop table query for the corresponding table
        """
        return {
            group: "DROP TABLE IF EXISTS %s;" % self.get_table_name(group)
            for group in self.groups
        }

    def get_indexes(self):
        """
        Generate create index queries for this aggregation

        Returns: a dictionary of group : index pairs where
            group are the same keys as groups
            index is a raw create index query for the corresponding table
        """
        return {
            group: "CREATE INDEX ON %s (%s);" % (self.get_table_name(group), groupby)
            for group, groupby in self.groups.items()
        }

    def get_join_table(self):
        """
        Generate a query for a join table
        """
        return ex.Select(
            columns=[make_sql_clause(group, ex.column) for group in self.groups.values()],
            from_obj=self.from_obj
        ).group_by(
            *self.groups.values()
        )

    def get_create(self, join_table=None):
        """
        Generate a single aggregation table creation query by joining
            together the results of get_creates()
        Returns: a CREATE TABLE AS query
        """
        if not join_table:
            join_table = "(%s) t1" % self.get_join_table()

        query = "SELECT * FROM %s\n" % join_table
        for group, groupby in self.groups.items():
            query += "LEFT JOIN %s USING (%s)" % (self.get_table_name(group), groupby)

        return "CREATE TABLE %s AS (%s);" % (self.get_table_name(), query)

    def get_drop(self, imputed=False):
        """
        Generate a drop table statement for the aggregation table
        Returns: string sql query
        """
        return "DROP TABLE IF EXISTS %s" % self.get_table_name(imputed=imputed)

    def get_create_schema(self):
        """
        Generate a create schema statement
        """
        if self.schema is not None:
            return "CREATE SCHEMA IF NOT EXISTS %s" % self.schema

    def find_nulls(self, imputed=False):
        """
        Generate query to count number of nulls in each column in the aggregation table

        Returns: a SQL SELECT statement
        """
        query_template = """
            SELECT {cols}
            FROM {state_tbl} t1
            LEFT JOIN {aggs_tbl} t2 USING({group})
            """
        cols_sql = ",\n".join(
            [
                """SUM(CASE WHEN "{col}" IS NULL THEN 1 ELSE 0 END) AS "{col}" """.format(
                    col=column
                )
                for column in self.get_imputation_rules().keys()
            ]
        )

        return query_template.format(
            cols=cols_sql,
            state_tbl=self.state_table,
            aggs_tbl=self.get_table_name(imputed=imputed),
            group=self.state_group,
        )

    def _get_impute_select(self, impute_cols, nonimpute_cols, partitionby=None):

        imprules = self.get_imputation_rules()

        # check if we're missing any columns relative to the full set and raise an
        # exception if we are
        missing_cols = set(imprules.keys()) - set(nonimpute_cols + impute_cols)
        if len(missing_cols) > 0:
            raise ValueError("Missing columns in get_impute_create: %s" % missing_cols)

        # key columns and date column
        query = ""

        used_impflags = set()
        # pre-sort and iterate through the combined set to ensure column order
        for col in sorted(nonimpute_cols + impute_cols):
            # just pass through columns that don't require imputation (no nulls found)
            if col in nonimpute_cols:
                query += '\n,"%s"' % col

            # for columns that do require imputation, include SQL to do the imputation work
            # and a flag for whether the value was imputed
            if col in impute_cols:

                # we don't want to add redundant imputation flags. for a given source
                # column and time interval, all of the functions will have identical
                # sets of rows that needed imputation
                # to reliably merge these, we lookup the original aggregate that produced
                # the function, and see its available functions. we expect exactly one of
                # these functions to end the column name and remove it if so
                # this is passed to the imputer
                if hasattr(self.colname_aggregate_lookup[col], 'functions'):
                    agg_functions = self.colname_aggregate_lookup[col].functions
                    used_function = next(funcname for funcname in agg_functions if col.endswith(funcname))
                    if used_function in AGGFUNCS_NEED_MULTIPLE_VALUES:
                        impflag_basecol = col
                    else:
                        impflag_basecol = col.rstrip('_' + used_function)
                else:
                    logger.warning("Imputation flag merging is not implemented for "
                                    "AggregateExpression objects that don't define an aggregate "
                                    "function (e.g. composites)")
                    impflag_basecol = col
                impute_rule = imprules[col]

                try:
                    imputer = available_imputations[impute_rule["type"]]
                except KeyError as err:
                    raise ValueError(
                        "Invalid imputation type %s for column %s"
                        % (impute_rule.get("type", ""), col)
                    ) from err

                imputer = imputer(column=col, column_base_for_impflag=impflag_basecol, partitionby=partitionby, **impute_rule)

                query += "\n,%s" % imputer.to_sql()
                if not imputer.noflag:
                    # Add an imputation flag for non-categorical columns (this is handeled
                    # for categorical columns with a separate NULL category)
                    # but only add it if another functionally equivalent impflag hasn't already been added
                    impflag_select, impflag_alias = imputer.imputed_flag_select_and_alias()
                    if impflag_alias not in used_impflags:
                        used_impflags.add(impflag_alias)
                        query += "\n,%s as \"%s\" " % (impflag_select, impflag_alias)

        return query

    def get_impute_create(self, impute_cols, nonimpute_cols):
        """
        Generates the CREATE TABLE query for the aggregation table with imputation.

        Args:
            impute_cols: a list of column names with null values
            nonimpute_cols: a list of column names without null values

        Returns: a CREATE TABLE AS query
        """

        # key columns and date column
        query = "SELECT %s" % ", ".join(map(str, self.groups.values()))

        # columns with imputation filling as needed
        query += self._get_impute_select(impute_cols, nonimpute_cols)

        # imputation starts from the state table and left joins into the aggregation table
        query += "\nFROM %s t1" % self.state_table
        query += "\nLEFT JOIN %s t2 USING(%s)" % (
            self.get_table_name(),
            self.state_group,
        )

        return "CREATE TABLE %s AS (%s)" % (self.get_table_name(imputed=True), query)

    def execute(self, conn, join_table=None):
        """
        Execute all SQL statements to create final aggregation table.
        Args:
            conn: the SQLAlchemy connection on which to execute
        """
        self.validate(conn)
        create_schema = self.get_create_schema()
        creates = self.get_creates()
        drops = self.get_drops()
        indexes = self.get_indexes()
        inserts = self.get_inserts()
        drop = self.get_drop()
        create = self.get_create(join_table=join_table)

        trans = conn.begin()

        if create_schema is not None:
            conn.execute(create_schema)

        for group in self.groups:
            conn.execute(drops[group])
            conn.execute(creates[group])
            for insert in inserts[group]:
                conn.execute(insert)
            conn.execute(indexes[group])

        # create the aggregation table
        conn.execute(drop)
        conn.execute(create)

        # excute query to find columns with null values and create lists of columns
        # that do and do not need imputation when creating the imputation table
        res = conn.execute(self.find_nulls())
        null_counts = list(zip(res.keys(), res.fetchone()))
        impute_cols = [col for col, val in null_counts if val > 0]
        nonimpute_cols = [col for col, val in null_counts if val == 0]
        res.close()

        # sql to drop and create the imputation table
        drop_imp = self.get_drop(imputed=True)
        create_imp = self.get_impute_create(
            impute_cols=impute_cols, nonimpute_cols=nonimpute_cols
        )

        # create the imputation table
        conn.execute(drop_imp)
        conn.execute(create_imp)

        trans.commit()

    def validate(self, conn):
        """
        Validate the Aggregation to ensure that it will perform as expected.
        This is done against an active SQL connection in order to enable
        validation of the SQL itself.
        """



import unittest
from triage.component.audition.selection_rules import BoundSelectionRule
from triage.component.audition.selection_rule_grid import make_selection_rule_grid

from triage.component.audition.rules_maker import (
    SimpleRuleMaker,
    RandomGroupRuleMaker,
    TwoMetricsRuleMaker,
    create_selection_grid,
)


class TestSimpleRuleMaker(unittest.TestCase):
    def test_add_rule_best_current_value(self):
        Rule = SimpleRuleMaker()
        Rule.add_rule_best_current_value(metric="precision@", parameter="100_abs")
        assert Rule.create() == [
            {
                "selection_rules": [{"name": "best_current_value", "n": 1}],
                "shared_parameters": [{"metric": "precision@", "parameter": "100_abs"}],
            }
        ]

    def test_add_rule_best_average_value(self):
        Rule = SimpleRuleMaker()
        Rule.add_rule_best_average_value(metric="precision@", parameter="100_abs")
        assert Rule.create() == [
            {
                "selection_rules": [{"name": "best_average_value", "n": 1}],
                "shared_parameters": [{"metric": "precision@", "parameter": "100_abs"}],
            }
        ]

    def test_add_rule_lowest_metric_variance(self):
        Rule = SimpleRuleMaker()
        Rule.add_rule_lowest_metric_variance(metric="precision@", parameter="100_abs")
        assert Rule.create() == [
            {
                "selection_rules": [{"name": "lowest_metric_variance", "n": 1}],
                "shared_parameters": [{"metric": "precision@", "parameter": "100_abs"}],
            }
        ]

    def test_add_rule_most_frequent_best_dist(self):
        Rule = SimpleRuleMaker()
        Rule.add_rule_most_frequent_best_dist(
            metric="precision@",
            parameter="100_abs",
            dist_from_best_case=[0.01, 0.05, 0.1, 0.15],
        )
        assert Rule.create() == [
            {
                "selection_rules": [
                    {
                        "dist_from_best_case": [0.01, 0.05, 0.1, 0.15],
                        "name": "most_frequent_best_dist",
                        "n": 1,
                    }
                ],
                "shared_parameters": [{"metric": "precision@", "parameter": "100_abs"}],
            }
        ]

    def test_add_rule_best_avg_recency_weight(self):
        Rule = SimpleRuleMaker()
        Rule.add_rule_best_avg_recency_weight(metric="precision@", parameter="100_abs")
        assert Rule.create() == [
            {
                "selection_rules": [
                    {
                        "curr_weight": [1.5, 2.0, 5.0],
                        "decay_type": ["linear"],
                        "name": "best_avg_recency_weight",
                        "n": 1,
                    }
                ],
                "shared_parameters": [{"metric": "precision@", "parameter": "100_abs"}],
            }
        ]

    def test_add_rule_best_avg_var_penalized(self):
        Rule = SimpleRuleMaker()
        Rule.add_rule_best_avg_var_penalized(
            metric="precision@", parameter="100_abs", stdev_penalty=0.5
        )
        assert Rule.create() == [
            {
                "selection_rules": [
                    {"name": "best_avg_var_penalized", "stdev_penalty": 0.5, "n": 1}
                ],
                "shared_parameters": [{"metric": "precision@", "parameter": "100_abs"}],
            }
        ]


class TestRandomGroupRuleMaker(unittest.TestCase):
    def test_random_model_groups(self):
        Rule = RandomGroupRuleMaker()
        assert Rule.create() == [
            {
                "selection_rules": [{"name": "random_model_group", "n": 1}],
                "shared_parameters": [{}],
            }
        ]


class TestTwoMetricsRuleMaker(unittest.TestCase):
    def test_add_two_metrics_rule_maker(self):
        Rule = TwoMetricsRuleMaker()
        Rule.add_rule_best_average_two_metrics(
            metric1="precision@",
            parameter1="100_abs",
            metric2="recall@",
            parameter2="300_abs",
            metric1_weight=[0.5],
        )
        assert Rule.create() == [
            {
                "selection_rules": [
                    {
                        "metric1_weight": [0.5],
                        "name": "best_average_two_metrics",
                        "metric2": ["recall@"],
                        "parameter2": ["300_abs"],
                        "n": 1,
                    }
                ],
                "shared_parameters": [
                    {"metric1": "precision@", "parameter1": "100_abs"}
                ],
            }
        ]


class TestCreateSelectionRuleGrid(unittest.TestCase):
    def test_create_grid(self):
        """
        input_data = [{
            'shared_parameters': [
                {'metric': 'precision@', 'parameter': '100_abs'},
                {'metric': 'recall@', 'parameter': '100_abs'},
            ],
            'selection_rules': [
                {'name': 'most_frequent_best_dist', 'dist_from_best_case': [0.1, 0.2, 0.3]},
                {'name': 'best_current_value'}
            ]
        }, {
            'shared_parameters': [
                {'metric1': 'precision@', 'parameter1': '100_abs'},
            ],
            'selection_rules': [
                {
                    'name': 'best_average_two_metrics',
                    'metric2': ['recall@'],
                    'parameter2': ['100_abs'],
                    'metric1_weight': [0.4, 0.5, 0.6]
                },
            ]
        }]
        """
        Rule1 = SimpleRuleMaker()
        Rule1.add_rule_best_current_value(metric="precision@", parameter="100_abs")
        Rule1.add_rule_most_frequent_best_dist(
            metric="recall@", parameter="100_abs", dist_from_best_case=[0.1, 0.2, 0.3]
        )

        Rule2 = TwoMetricsRuleMaker()
        Rule2.add_rule_best_average_two_metrics(
            metric1="precision@",
            parameter1="100_abs",
            metric2="recall@",
            parameter2="100_abs",
            metric1_weight=[0.4, 0.5, 0.6],
        )

        expected_output = [
            BoundSelectionRule(
                descriptive_name="most_frequent_best_dist_precision@_100_abs_0.1",
                function_name="most_frequent_best_dist",
                args={
                    "metric": "precision@",
                    "parameter": "100_abs",
                    "dist_from_best_case": 0.1,
                },
            ),
            BoundSelectionRule(
                descriptive_name="most_frequent_best_dist_precision@_100_abs_0.2",
                function_name="most_frequent_best_dist",
                args={
                    "metric": "precision@",
                    "parameter": "100_abs",
                    "dist_from_best_case": 0.2,
                },
            ),
            BoundSelectionRule(
                descriptive_name="most_frequent_best_dist_precision@_100_abs_0.3",
                function_name="most_frequent_best_dist",
                args={
                    "metric": "precision@",
                    "parameter": "100_abs",
                    "dist_from_best_case": 0.3,
                },
            ),
            BoundSelectionRule(
                descriptive_name="most_frequent_best_dist_recall@_100_abs_0.1",
                function_name="most_frequent_best_dist",
                args={
                    "metric": "recall@",
                    "parameter": "100_abs",
                    "dist_from_best_case": 0.1,
                },
            ),
            BoundSelectionRule(
                descriptive_name="most_frequent_best_dist_recall@_100_abs_0.2",
                function_name="most_frequent_best_dist",
                args={
                    "metric": "recall@",
                    "parameter": "100_abs",
                    "dist_from_best_case": 0.2,
                },
            ),
            BoundSelectionRule(
                descriptive_name="most_frequent_best_dist_recall@_100_abs_0.3",
                function_name="most_frequent_best_dist",
                args={
                    "metric": "recall@",
                    "parameter": "100_abs",
                    "dist_from_best_case": 0.3,
                },
            ),
            BoundSelectionRule(
                descriptive_name="best_current_value_precision@_100_abs",
                function_name="best_current_value",
                args={"metric": "precision@", "parameter": "100_abs"},
            ),
            BoundSelectionRule(
                descriptive_name="best_current_value_recall@_100_abs",
                function_name="best_current_value",
                args={"metric": "recall@", "parameter": "100_abs"},
            ),
            BoundSelectionRule(
                descriptive_name="best_average_two_metrics_precision@_100_abs_recall@_100_abs_0.4",
                function_name="best_average_two_metrics",
                args={
                    "metric1": "precision@",
                    "parameter1": "100_abs",
                    "metric2": "recall@",
                    "parameter2": "100_abs",
                    "metric1_weight": 0.4,
                },
            ),
            BoundSelectionRule(
                descriptive_name="best_average_two_metrics_precision@_100_abs_recall@_100_abs_0.5",
                function_name="best_average_two_metrics",
                args={
                    "metric1": "precision@",
                    "parameter1": "100_abs",
                    "metric2": "recall@",
                    "parameter2": "100_abs",
                    "metric1_weight": 0.5,
                },
            ),
            BoundSelectionRule(
                descriptive_name="best_average_two_metrics_precision@_100_abs_recall@_100_abs_0.6",
                function_name="best_average_two_metrics",
                args={
                    "metric1": "precision@",
                    "parameter1": "100_abs",
                    "metric2": "recall@",
                    "parameter2": "100_abs",
                    "metric1_weight": 0.6,
                },
            ),
        ]
        expected_output.sort(key=lambda x: x.descriptive_name)
        grid = sorted(
            make_selection_rule_grid(create_selection_grid(Rule1, Rule2)),
            key=lambda x: x.descriptive_name,
        )
        assert len(grid) == len(expected_output)
        for expected_rule, actual_rule in zip(expected_output, grid):
            assert expected_rule.descriptive_name == actual_rule.descriptive_name



from sqlalchemy import create_engine
import testing.postgresql
import pytest

from triage.component.catwalk.db import ensure_db

from tests.utils import sample_config, populate_source_data
from triage.experiments.defaults import (
    fill_timechop_config_missing,
    fill_cohort_config_missing,
    fill_feature_group_definition,
    fill_model_grid_presets,
    model_grid_preset,
)


def test_fill_cohort_config_missing():
    config = sample_config()
    config.pop('cohort_config')
    cohort_config = fill_cohort_config_missing(config)
    assert cohort_config == {
        'query': "select distinct entity_id from "
                "((select entity_id, as_of_date as knowledge_date from "
                "(select * from cat_complaints) as t)\n union \n(select entity_id, "
                "as_of_date as knowledge_date from (select * from entity_zip_codes "
                "join zip_code_events using (zip_code)) as t)) as e "
                "where knowledge_date < '{as_of_date}'",
        'name': 'all_entities'
        }


def test_fill_feature_group_definition():
    config = sample_config()
    fg_definition = fill_feature_group_definition(config)
    assert sorted(fg_definition['prefix']) == ['entity_features', 'zip_code_features']


def test_fill_timechop_config_missing():
    remove_keys = [
        'model_update_frequency',
        'training_as_of_date_frequencies',
        'test_as_of_date_frequencies',
        'max_training_histories',
        'test_durations',
        'feature_start_time',
        'feature_end_time',
        'label_start_time',
        'label_end_time',
        'training_label_timespans',
        'test_label_timespans'
        ]

    # ensure redundant keys properly raise errors
    config = sample_config()
    config['temporal_config']['label_timespans'] = '1y'
    with pytest.raises(KeyError):
        timechop_config = fill_timechop_config_missing(config, None)

    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        ensure_db(db_engine)
        populate_source_data(db_engine)
        config = sample_config()

        for key in remove_keys:
            config['temporal_config'].pop(key)
        config['temporal_config']['label_timespans'] = '1y'

        timechop_config = fill_timechop_config_missing(config, db_engine)

        assert timechop_config['model_update_frequency'] == '100y'
        assert timechop_config['training_as_of_date_frequencies'] == '100y'
        assert timechop_config['test_as_of_date_frequencies'] == '100y'
        assert timechop_config['max_training_histories'] == '0d'
        assert timechop_config['test_durations'] == '0d'
        assert timechop_config['training_label_timespans'] == '1y'
        assert timechop_config['test_label_timespans'] == '1y'
        assert 'label_timespans' not in timechop_config.keys()
        assert timechop_config['feature_start_time'] == '2010-10-01'
        assert timechop_config['feature_end_time'] == '2013-10-01'
        assert timechop_config['label_start_time'] == '2010-10-01'
        assert timechop_config['label_end_time'] == '2013-10-01'


def test_model_grid_preset():
    test_cases = {
        'quickstart': {'total_items': 8, 'model_types': 3},
        'small': {'total_items': 35, 'model_types': 4},
        'medium': {'total_items': 64, 'model_types': 6},
        'large': {'total_items': 83, 'model_types': 6},
        'texas': {'total_items': 115, 'model_types': 9},
    }

    for grid_type, exp_results in test_cases.items():
        preset_grid = model_grid_preset(grid_type)
        assert len(preset_grid) == exp_results['model_types']
        total_items = sum([len(y) for x in preset_grid.values() for y in x.values()])
        assert total_items == exp_results['total_items']


def test_fill_model_grid_presets():

    # case 1: has grid, no preset
    config = sample_config()
    fill_grid = fill_model_grid_presets(config)
    assert fill_grid == config['grid_config']

    # case 2: has preset, no grid
    config = sample_config()
    config.pop('grid_config')
    config['model_grid_preset'] = 'quickstart'
    fill_grid = fill_model_grid_presets(config)
    assert len(fill_grid) == 3

    # case 3: neither
    config = sample_config()
    config.pop('grid_config')
    fill_grid = fill_model_grid_presets(config)
    assert fill_grid is None

    # case 4: both
    config = sample_config()
    config['model_grid_preset'] = 'quickstart'
    fill_grid = fill_model_grid_presets(config)
    assert len(fill_grid) == 3
    assert len(fill_grid.get('sklearn.tree.DecisionTreeClassifier', {}).get('max_depth', [])) == 3
    assert len(fill_grid.get('sklearn.tree.DecisionTreeClassifier', {}).get('criterion', [])) == 1



import os
from os.path import isfile, join
from tempfile import TemporaryDirectory

import testing.postgresql
from triage import create_engine


from tests.utils import sample_config, populate_source_data

from triage.experiments import SingleThreadedExperiment
from triage.database_reflection import schema_tables
from triage.validation_primitives import table_should_have_data
from unittest import TestCase
from contextlib import contextmanager


@contextmanager
def prepare_experiment(config):
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        populate_source_data(db_engine)
        with TemporaryDirectory() as temp_dir:
            experiment = SingleThreadedExperiment(
                config=config,
                db_engine=db_engine,
                project_path=os.path.join(temp_dir, "inspections"),
                cleanup=False,
                partial_run=True,
            )
            yield experiment


class GetSplits(TestCase):
    config = {
        "temporal_config": sample_config()["temporal_config"],
        "config_version": sample_config()["config_version"],
        "random_seed": sample_config()["random_seed"],
    }

    def test_run(self):
        with prepare_experiment(self.config) as experiment:
            experiment.run()
            assert experiment.split_definitions

    def test_validate_nonstrict(self):
        with prepare_experiment(self.config) as experiment:
            experiment.validate(strict=False)

    def test_validate_strict(self):
        with prepare_experiment(self.config) as experiment:
            with self.assertRaises(ValueError):
                experiment.validate()


class Cohort(TestCase):
    config = {
        "temporal_config": sample_config()["temporal_config"],
        "cohort_config": sample_config()["cohort_config"],
        "config_version": sample_config()["config_version"],
        "random_seed": sample_config()["random_seed"],
    }

    def test_run(self):
        with prepare_experiment(self.config) as experiment:
            experiment.run()
            table_should_have_data(
                experiment.cohort_table_name, experiment.db_engine
            )

    def test_validate_nonstrict(self):
        with prepare_experiment(self.config) as experiment:
            experiment.validate(strict=False)

    def test_validate_strict(self):
        with prepare_experiment(self.config) as experiment:
            with self.assertRaises(ValueError):
                experiment.validate()


class Labels(TestCase):
    config = {
        "temporal_config": sample_config()["temporal_config"],
        "label_config": sample_config()["label_config"],
        "config_version": sample_config()["config_version"],
        "random_seed": sample_config()["random_seed"],
    }

    def test_run(self):
        with prepare_experiment(self.config) as experiment:
            experiment.run()
            table_should_have_data(experiment.labels_table_name, experiment.db_engine)

    def test_validate_nonstrict(self):
        with prepare_experiment(self.config) as experiment:
            experiment.validate(strict=False)

    def test_validate_strict(self):
        with prepare_experiment(self.config) as experiment:
            with self.assertRaises(ValueError):
                experiment.validate()


class PreimputationFeatures(TestCase):
    config = {
        "temporal_config": sample_config()["temporal_config"],
        "feature_aggregations": sample_config()["feature_aggregations"],
        "config_version": sample_config()["config_version"],
        "random_seed": sample_config()["random_seed"],
    }

    def test_run(self):
        with prepare_experiment(self.config) as experiment:
            experiment.run()
            generated_tables = [
                table
                for table in schema_tables(
                    experiment.features_schema_name, experiment.db_engine
                ).keys()
                if "_aggregation" in table
            ]

            assert len(generated_tables) == len(sample_config()["feature_aggregations"])
            for table in generated_tables:
                table_should_have_data(table, experiment.db_engine)

    def test_validate_nonstrict(self):
        with prepare_experiment(self.config) as experiment:
            experiment.validate(strict=False)

    def test_validate_strict(self):
        with prepare_experiment(self.config) as experiment:
            with self.assertRaises(ValueError):
                experiment.validate()


class PostimputationFeatures(TestCase):
    config = {
        "temporal_config": sample_config()["temporal_config"],
        "feature_aggregations": sample_config()["feature_aggregations"],
        "cohort_config": sample_config()["cohort_config"],
        "config_version": sample_config()["config_version"],
        "random_seed": sample_config()["random_seed"],
    }

    def test_run(self):
        with prepare_experiment(self.config) as experiment:
            experiment.run()
            generated_tables = [
                table
                for table in schema_tables(
                    experiment.features_schema_name, experiment.db_engine
                ).keys()
                if "_aggregation_imputed" in table
            ]

            assert len(generated_tables) == len(sample_config()["feature_aggregations"])
            for table in generated_tables:
                table_should_have_data(table, experiment.db_engine)

    def test_validate_nonstrict(self):
        with prepare_experiment(self.config) as experiment:
            experiment.validate(strict=False)

    def test_validate_strict(self):
        with prepare_experiment(self.config) as experiment:
            with self.assertRaises(ValueError):
                experiment.validate()


class Matrices(TestCase):
    config = {
        "temporal_config": sample_config()["temporal_config"],
        "feature_aggregations": sample_config()["feature_aggregations"],
        "cohort_config": sample_config()["cohort_config"],
        "label_config": sample_config()["label_config"],
        "config_version": sample_config()["config_version"],
        "random_seed": sample_config()["random_seed"],
    }

    def test_run(self):
        with prepare_experiment(self.config) as experiment:
            experiment.run()
            matrices_path = join(experiment.project_path, "matrices")
            matrices_and_metadata = [
                f for f in os.listdir(matrices_path) if isfile(join(matrices_path, f))
            ]
            matrices = experiment.matrix_build_tasks
            assert len(matrices) > 0
            for matrix in matrices:
                assert "{}.csv.gz".format(matrix) in matrices_and_metadata
                assert "{}.yaml".format(matrix) in matrices_and_metadata

    def test_validate_nonstrict(self):
        with prepare_experiment(self.config) as experiment:
            experiment.validate(strict=False)

    def test_validate_strict(self):
        with prepare_experiment(self.config) as experiment:
            with self.assertRaises(ValueError):
                experiment.validate()



"""
Postmodeling parameters

This script contain the parameters Class that will be used across all the
postmodeling functions within the ModelEvaluator and ModelGroupEvaluator
classes. This class will be initialized using the 'postmodeling_parameters.yaml'
file

"""

import yaml
import json

import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

class PostmodelParameters:
    '''
    PostmodelParameters reads all parameters from a 'yaml' file and store them
    in this object to be used in other functions. Different metrics can be
    passed to this object, by default it will reads from a
    'postmodeling_parameters.yaml', but an Audition config file can be passed
    and will parse from it the needed parameters
    '''
    def __init__(self, path_params):

        with open(path_params) as f:
            params = yaml.full_load(f)

        # Assign dict elements to Parameters object and flatten
        # thresholds
        self.__dict__.update(params)
        self.figsize = tuple(self.figsize)

        try:
            if self.audition_output_path is not None:
                with open(self.audition_output_path) as f:
                    json_models = json.load(f)

                list_models = [model for model_list in json_models.values()
                               for model in model_list]
                self.model_group_id = list_models

        except AttributeError:
            logger.exception(
                f'''No audition output file was defined. I will use the models
                defined in the {path_params} configuration file.'''
            )



import datetime
import random
import tempfile
from contextlib import contextmanager
import pytest

import numpy as np
import pandas as pd
import yaml

from triage.component.catwalk.storage import (
    ProjectStorage,
)
from triage.util.structs import FeatureNameList


def fake_labels(length):
    return np.array([random.choice([True, False]) for i in range(0, length)])


@pytest.fixture
def sample_metadata():
    return {
        "feature_start_time": datetime.date(2012, 12, 20),
        "end_time": datetime.date(2016, 12, 20),
        "label_name": "label",
        "as_of_date_frequency": "1w",
        "max_training_history": "5y",
        "state": "default",
        "cohort_name": "default",
        "label_timespan": "1y",
        "metta-uuid": "1234",
        "feature_names": FeatureNameList(["ft1", "ft2"]),
        "feature_groups": ["all: True"],
        "indices": ["entity_id"],
    }


@pytest.fixture
def sample_df():
    return pd.DataFrame.from_dict(
        {
            "entity_id": [1, 2],
            "feature_one": [3, 4],
            "feature_two": [5, 6],
            "label": ["good", "bad"],
        }
    ).set_index("entity_id")


@pytest.fixture
def sample_matrix_store(sample_df, sample_metadata):
    with tempfile.TemporaryDirectory() as tempdir:
        project_storage = ProjectStorage(tempdir)
        store = project_storage.matrix_storage_engine().get_store("1234")
        store.matrix = sample_df
        store.metadata = sample_metadata
        return store



from triage.component.architect.feature_generators import FeatureGenerator
from triage.component.architect.feature_dictionary_creator import (
    FeatureDictionaryCreator,
)
from triage.component.architect.feature_group_creator import FeatureGroupCreator
from triage.component.architect.feature_group_mixer import FeatureGroupMixer

__all__ = (
    "FeatureGenerator",
    "FeatureDictionaryCreator",
    "FeatureGroupCreator",
    "FeatureGroupMixer",
)



"""hash-partitioning predictions tables

Revision ID: 45219f25072b
Revises: a98acf92fd48
Create Date: 2020-08-21 09:29:04.751933

"""
from alembic import op
import os

import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)


# revision identifiers, used by Alembic.
revision = '45219f25072b'
down_revision = 'a98acf92fd48'
branch_labels = None
depends_on = None


def get_pg_major_version(op):
    conn = op.get_bind()
    pg_major_version = conn.execute('show server_version').fetchone()[0].split('.')[0]
    logger.debug(f'PostgreSQL major version {pg_major_version}')
    return int(pg_major_version)


def upgrade():

    pg_major_version = get_pg_major_version(op)

    if pg_major_version >= 11:
        logger.info(f'PostgreSQL 11 or greater found (PostgreSQL {pg_major_version}): Using hash partitioning')
        hash_partitioning_filename = os.path.join(
            os.path.dirname(__file__), "../../sql/predictions_hash_partitioning.sql"
        )
        with open(hash_partitioning_filename) as fd:
            stmt = fd.read()
            op.execute(stmt)
    else:
        logger.info(f'No hash partitioning implemented because PostgreSQL 11 or greater not found (using: PostgreSQL {pg_major_version})')


def downgrade():

    pg_major_version = get_pg_major_version(op)

    if pg_major_version >= 11:
        logger.info(f'PostgreSQL 11 or greater found  (PostgreSQL {pg_major_version}): Removing hash partitioning')
        undo_hash_partitioning_filename = os.path.join(
            os.path.dirname(__file__), "../../sql/undo_predictions_hash_partitioning.sql"
        )
        with open(undo_hash_partitioning_filename) as fd:
            stmt = fd.read()
            op.execute(stmt)
    else:
        logger.info(f'No hash partitioning implemented because PostgreSQL 11 or greater not found (using: PostgreSQL {pg_major_version})')



import matplotlib
matplotlib.use('Agg')
import matplotlib.dates as md
import numpy as np
from triage.util.conf import convert_str_to_relativedelta
import matplotlib.pyplot as plt


FIG_SIZE = (32, 16)


def visualize_chops(chopper, show_as_of_times=True, show_boundaries=True, save_target=None):
    """Visualize time chops of a given Timechop object using matplotlib

    Args:
        chopper (triage.component.timechop.Timechop): A fully-configured Timechop object
        show_as_of_times (bool): Whether or not to draw horizontal lines
            for as-of-times
        show_boundaries (bool): Whether or not to show a rectangle around matrices
            and dashed lines around feature/label boundaries
        save_target (str or file-like object): A save target for matplotlib to save
            the figure to. Defaults to None, which won't save anything
    """
    chops = chopper.chop_time()

    chops.reverse()

    fig, ax = plt.subplots(nrows=len(chops), sharex=True, sharey=True, squeeze=False, figsize=FIG_SIZE)

    for idx, chop in enumerate(chops):
        train_as_of_times = chop["train_matrix"]["as_of_times"]
        test_as_of_times = chop["test_matrices"][0]["as_of_times"]

        test_label_timespan = chop["test_matrices"][0]["test_label_timespan"]
        training_label_timespan = chop["train_matrix"]["training_label_timespan"]

        color_rgb = np.random.random(3)

        if show_as_of_times:
            # Train matrix (as_of_times)
            ax[idx][0].hlines(
                [x for x in range(len(train_as_of_times))],
                [x.date() for x in train_as_of_times],
                [
                    x.date() + convert_str_to_relativedelta(training_label_timespan)
                    for x in train_as_of_times
                ],
                linewidth=3,
                color=color_rgb,
                label=f"train_{idx}",
            )

            # Test matrix
            ax[idx][0].hlines(
                [x for x in range(len(test_as_of_times))],
                [x.date() for x in test_as_of_times],
                [
                    x.date() + convert_str_to_relativedelta(test_label_timespan)
                    for x in test_as_of_times
                ],
                linewidth=3,
                color=color_rgb,
                label=f"test_{idx}",
            )

        if show_boundaries:
            # Limits: train
            ax[idx][0].axvspan(
                chop["train_matrix"]["first_as_of_time"],
                chop["train_matrix"]["last_as_of_time"],
                color=color_rgb,
                alpha=0.3,
            )

            ax[idx][0].axvline(
                chop["train_matrix"]["matrix_info_end_time"], color="k", linestyle="--"
            )

            # Limits: test
            ax[idx][0].axvspan(
                chop["test_matrices"][0]["first_as_of_time"],
                chop["test_matrices"][0]["last_as_of_time"],
                color=color_rgb,
                alpha=0.3,
            )

            ax[idx][0].axvline(
                chop["feature_start_time"], color="k", linestyle="--", alpha=0.2
            )
            ax[idx][0].axvline(
                chop["feature_end_time"], color="k", linestyle="--", alpha=0.2
            )
            ax[idx][0].axvline(
                chop["label_start_time"], color="k", linestyle="--", alpha=0.2
            )
            ax[idx][0].axvline(
                chop["label_end_time"], color="k", linestyle="--", alpha=0.2
            )

            ax[idx][0].axvline(
                chop["test_matrices"][0]["matrix_info_end_time"],
                color="k",
                linestyle="--",
            )

        ax[idx][0].yaxis.set_major_locator(plt.NullLocator())
        ax[idx][0].yaxis.set_label_position("right")
        ax[idx][0].set_ylabel(f'Label timespan \n {test_label_timespan} (test), {training_label_timespan} (training)',
                              rotation="vertical", labelpad=30)

        ax[idx][0].xaxis.set_major_formatter(md.DateFormatter("%Y"))
        ax[idx][0].xaxis.set_major_locator(md.YearLocator())
        ax[idx][0].xaxis.set_minor_locator(md.MonthLocator())

    ax[0][0].set_title("Timechop: Temporal cross-validation blocks")
    fig.subplots_adjust(hspace=0)
    plt.setp([a.get_xticklabels() for a in fig.axes[:-1]], visible=False)
    if save_target:
        plt.savefig(save_target)
    plt.show()



from triage.component.timechop.utils import convert_to_list


def test_convert_to_list():
    tests = [
        {"val": "1 day", "expected_result": ["1 day"]},
        {"val": ["1 day"], "expected_result": ["1 day"]},
        {"val": 1, "expected_result": [1]},
    ]
    for test in tests:
        assert convert_to_list(test["val"]) == test["expected_result"]



import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array

class CutOff(BaseEstimator, TransformerMixin):
    """Transform features cutting values out of established range

    Args:
       feature_range: Range of allowed values, default=`(0,1)`

    Usage:
       The recommended way of using this is::

           from sklearn.pipeline import Pipeline

           minmax_scaler = preprocessing.MinMaxScaler()
           dsapp_cutoff = CutOff()
           lr  = linear_model.LogisticRegression()

           pipeline =Pipeline([
                 ('minmax_scaler',minmax_scaler),
                 ('dsapp_cutoff', dsapp_cutoff),
                 ('lr', lr)
           ])

           pipeline.fit(X_train, y_train)
           pipeline.predict(X_test)

    """

    def __init__(self, feature_range=(0, 1), copy=True):
        self.feature_range = feature_range
        self.copy = copy


    def fit(self, X, y=None):
        return self


    def transform(self, X):
        feature_range = self.feature_range

        X = check_array(X, copy=self.copy, ensure_2d=True)

        if np.any(X > feature_range[1]) or np.any(X < feature_range[0]):
            logger.notice(
                f"You got feature values that are out of the range: {feature_range}. "
                f"The feature values will cutoff to fit in the range {feature_range}."
            )

        X[X > feature_range[1]] = feature_range[1]
        X[X < feature_range[0]] = feature_range[0]

        return X



"""empty message

Revision ID: fa1760d35710
Revises: a20104116533
Create Date: 2020-07-16 18:07:58.229213

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'fa1760d35710'
down_revision = 'a20104116533'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('experiments', sa.Column('random_seed', sa.Integer(), nullable=True), schema='triage_metadata')
    # ### end Alembic commands ###


def downgrade():
    op.drop_column('experiments', 'random_seed', schema='triage_metadata')
    # ### end Alembic commands ###



"""empty message

Revision ID: 8cef808549dd
Revises: b4d7569d31cb
Create Date: 2020-06-02 21:26:32.528991

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '8cef808549dd'
down_revision = 'b4d7569d31cb'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('experiment_runs', sa.Column('python_version', sa.String(), nullable=True), schema='model_metadata')
    op.create_index(op.f('ix_model_metadata_models_model_hash'), 'models', ['model_hash'], unique=True, schema='model_metadata')
    op.drop_index('ix_results_models_model_hash', table_name='models', schema='model_metadata')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index('ix_results_models_model_hash', 'models', ['model_hash'], unique=True, schema='model_metadata')
    op.drop_index(op.f('ix_model_metadata_models_model_hash'), table_name='models', schema='model_metadata')
    op.drop_column('experiment_runs', 'python_version', schema='model_metadata')
    # ### end Alembic commands ###



"""Testing Factories for creating database objects in unit tests

If init_engine is called first, objects are instantiated in the session at
module level. Example:

```
from .factories import EvaluationFactory, session, init_engine

engine = # your engine creation code here
init_engine(engine)
EvaluationFactory()
session.commit()

results = engine.execute('select * from results.evaluations')
```

"""
from datetime import datetime

import factory
import factory.fuzzy
from sqlalchemy.orm import sessionmaker, scoped_session

from triage.component import results_schema as schema


sessionmaker = sessionmaker()
session = scoped_session(sessionmaker)


class ExperimentFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.Experiment
        sqlalchemy_session = session

    experiment_hash = factory.fuzzy.FuzzyText()
    config = {}


class ModelGroupFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.ModelGroup
        sqlalchemy_session = session

    model_type = "sklearn.ensemble.RandomForestClassifier"
    hyperparameters = {"hyperparam1": "value1", "hyperparam2": "value2"}
    feature_list = ["feature1", "feature2", "feature3"]
    model_config = {}


class MatrixFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.Matrix
        sqlalchemy_session = session

    matrix_uuid = factory.fuzzy.FuzzyText()


class BaseModelFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.Model
        sqlalchemy_session = session

    model_group_rel = factory.SubFactory(ModelGroupFactory)
    model_hash = factory.fuzzy.FuzzyText()
    run_time = factory.LazyFunction(lambda: datetime.now())
    batch_run_time = factory.LazyFunction(lambda: datetime.now())
    model_type = "sklearn.ensemble.RandomForestClassifier"
    hyperparameters = {"hyperparam1": "value1", "hyperparam2": "value2"}
    model_comment = ""
    batch_comment = ""
    config = {}
    matrix_rel = factory.SubFactory(MatrixFactory)
    train_end_time = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))
    test = False
    train_matrix_uuid = factory.SelfAttribute("matrix_rel.matrix_uuid")
    training_label_timespan = "1y"


class ExperimentModelFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.ExperimentModel
        sqlalchemy_session = session

    model_rel = factory.SubFactory(BaseModelFactory)
    experiment_rel = factory.SubFactory(ExperimentFactory)


class ModelFactory(BaseModelFactory):
    experiment_association = factory.RelatedFactory(ExperimentModelFactory, 'model_rel')


class SubsetFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.Subset
        sqlalchemy_session = session

    subset_hash = factory.fuzzy.FuzzyText()
    config = {}
    created_timestamp = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))


class FeatureImportanceFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.FeatureImportance
        sqlalchemy_session = session

    model_rel = factory.SubFactory(ModelFactory)
    feature = factory.fuzzy.FuzzyText()
    feature_importance = factory.fuzzy.FuzzyDecimal(0, 1)
    rank_abs = 1
    rank_pct = 1.0


class PredictionFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.TestPrediction
        sqlalchemy_session = session

    model_rel = factory.SubFactory(ModelFactory)
    entity_id = factory.fuzzy.FuzzyInteger(0)
    as_of_date = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))
    score = factory.fuzzy.FuzzyDecimal(0, 1)
    label_value = factory.fuzzy.FuzzyInteger(0, 1)
    matrix_rel = factory.SubFactory(MatrixFactory)
    test_label_timespan = "3m"



class ListPredictionFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.ListPrediction
        sqlalchemy_session = session

    model_rel = factory.SubFactory(ModelFactory)
    entity_id = factory.fuzzy.FuzzyInteger(0)
    as_of_date = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))
    score = factory.fuzzy.FuzzyDecimal(0, 1)
    rank_abs = 1
    rank_pct = 1.0
    matrix_uuid = "efgh"
    test_label_timespan = "3m"


class IndividualImportanceFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.IndividualImportance
        sqlalchemy_session = session

    model_rel = factory.SubFactory(ModelFactory)
    entity_id = factory.fuzzy.FuzzyInteger(0)
    as_of_date = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))
    feature = factory.fuzzy.FuzzyText()
    feature_value = factory.fuzzy.FuzzyDecimal(0, 100)
    method = factory.fuzzy.FuzzyText()
    importance_score = factory.fuzzy.FuzzyDecimal(0, 1)


class EvaluationFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.TestEvaluation
        sqlalchemy_session = session

    model_rel = factory.SubFactory(ModelFactory)
    subset_hash = ''
    evaluation_start_time = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))
    evaluation_end_time = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))
    as_of_date_frequency = "3d"
    metric = "precision@"
    parameter = "100_abs"
    num_labeled_examples = 10
    num_labeled_above_threshold = 8
    num_positive_labels = 5
    sort_seed = 8
    best_value = factory.fuzzy.FuzzyDecimal(0, 1)
    worst_value = factory.fuzzy.FuzzyDecimal(0, 1)
    stochastic_value = factory.fuzzy.FuzzyDecimal(0, 1)
    num_sort_trials = 5
    standard_deviation = 0.05
    matrix_rel = factory.SubFactory(MatrixFactory)
    matrix_uuid = factory.SelfAttribute("matrix_rel.matrix_uuid")


class ExperimentRunFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = schema.ExperimentRun
        sqlalchemy_session = session

    experiment_rel = factory.SubFactory(ExperimentFactory)

    start_time = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))
    start_method = "run"
    git_hash = "abcd"
    triage_version = "4.2.0"
    python_version = '3.6.2 (default, May 28 2020, 13:23:43) \n[GCC 9.3.0]'
    platform = "Linux!!!"
    os_user = "dsapp"
    working_directory = "/the/best/directory"
    ec2_instance_type = "x2.128xlarge"
    log_location = "/the/logs"
    experiment_class_path = "triage.experiments.singlethreaded.SingleThreadedExperiment"
    experiment_kwargs = {}
    installed_libraries = ['triage']
    matrix_building_started = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))
    matrices_made = 0
    matrices_skipped = 0
    matrices_errored = 0
    model_building_started = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))
    models_made = 0
    models_skipped = 0
    models_errored = 0
    last_updated_time = factory.fuzzy.FuzzyNaiveDateTime(datetime(2008, 1, 1))
    current_status = schema.ExperimentRunStatus.started
    stacktrace = ""


def init_engine(new_engine):
    global sessionmaker, engine, session
    engine = new_engine
    session.remove()
    sessionmaker.configure(bind=engine)



import sqlalchemy.sql.expression as ex
from sqlalchemy.ext.compiler import compiles


def make_sql_clause(s, constructor):
    if not isinstance(s, ex.ClauseElement):
        return constructor(s)
    else:
        return s


class CreateTableAs(ex.Executable, ex.ClauseElement):
    def __init__(self, name, query):
        self.name = name
        self.query = query


@compiles(CreateTableAs)
def _create_table_as(element, compiler, **kw):
    return "CREATE TABLE %s AS %s" % (element.name, compiler.process(element.query))


class InsertFromSelect(ex.Executable, ex.ClauseElement):
    def __init__(self, name, query):
        self.name = name
        self.query = query


@compiles(InsertFromSelect)
def _insert_from_select(element, compiler, **kw):
    return "INSERT INTO %s (%s)" % (element.name, compiler.process(element.query))


def to_sql_name(name):
    return name.replace('"', "")



# -*- coding: utf-8 -*-
"""Integration tests for `collate` module."""
import testing.postgresql
from sqlalchemy import create_engine
from sqlalchemy.sql import expression as ex

from triage.component.collate import Aggregation, Aggregate
from triage.component.collate.spacetime import SpacetimeAggregation

from . import initialize_db


IMPUTE_RULES = {
    "coltype": "aggregate",
    "count": {"type": "mean"},
    "mode": {"type": "mean"},
}

Postgresql = testing.postgresql.PostgresqlFactory(
    cache_initialized_db=True, on_initialized=initialize_db.handler
)


def teardown_module():
    Postgresql.clear_cache()


def test_engine():
    with Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        ((result,),) = engine.execute("SELECT COUNT(*) FROM food_inspections")
    assert result == 966


def test_st_explicit_execute():
    agg = Aggregate({"F": "results='Fail'"}, ["count"], IMPUTE_RULES)
    mode = Aggregate("", "mode", IMPUTE_RULES, order="zip")
    st = SpacetimeAggregation(
        [agg, agg + agg, mode],
        from_obj=ex.table("food_inspections"),
        groups={"license": ex.column("license_no"), "zip": ex.column("zip")},
        intervals={"license": ["1 year", "2 years", "all"], "zip": ["1 year"]},
        dates=["2016-08-30", "2015-11-06"],
        state_table="inspection_states",
        state_group="license_no",
        date_column="inspection_date",
        prefix="food_inspections",
    )
    with Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        st.execute(engine.connect())


def test_st_lazy_execute():
    agg = Aggregate("results='Fail'", ["count"], IMPUTE_RULES)
    st = SpacetimeAggregation(
        [agg],
        from_obj="food_inspections",
        groups=["license_no", "zip"],
        intervals={"license_no": ["1 year", "2 years", "all"], "zip": ["1 year"]},
        dates=["2016-08-30", "2015-11-06"],
        state_table="inspection_states",
        state_group="license_no",
        date_column='"inspection_date"',
    )
    with Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        st.execute(engine.connect())


def test_st_execute_broadcast_intervals():
    agg = Aggregate("results='Fail'", ["count"], IMPUTE_RULES)
    st = SpacetimeAggregation(
        [agg],
        from_obj="food_inspections",
        groups=["license_no", "zip"],
        intervals=["1 year", "2 years", "all"],
        dates=["2016-08-30", "2015-11-06"],
        state_table="inspection_states",
        state_group="license_no",
        date_column='"inspection_date"',
    )
    with Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        st.execute(engine.connect())


def test_execute():
    agg = Aggregate("results='Fail'", ["count"], IMPUTE_RULES)
    st = Aggregation(
        [agg],
        from_obj="food_inspections",
        groups=["license_no", "zip"],
        state_table="all_licenses",
        state_group="license_no",
    )
    with Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        st.execute(engine.connect())


def test_execute_schema_output_date_column():
    agg = Aggregate("results='Fail'", ["count"], IMPUTE_RULES)
    st = SpacetimeAggregation(
        [agg],
        from_obj="food_inspections",
        groups=["license_no", "zip"],
        intervals={"license_no": ["1 year", "2 years", "all"], "zip": ["1 year"]},
        dates=["2016-08-30", "2015-11-06"],
        state_table="inspection_states_diff_colname",
        state_group="license_no",
        schema="agg",
        date_column='"inspection_date"',
        output_date_column="aggregation_date",
    )
    with Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        st.execute(engine.connect())



import argcmdr
import triage.cli as cli
from unittest.mock import Mock, patch
import os


# we do not need a real database URL but one SQLalchemy thinks looks like a real one
@patch.dict(os.environ, {'DATABASE_URL': 'postgresql://postgres@8.8.8.8/test'})
def try_command(*argv):
    try:
        argcmdr.main(cli.Triage, argv=('--tb',) + argv)
    except SystemExit as exc:
        raise AssertionError(exc)


def test_cli_singlethreadedexperiment():
    with patch('triage.cli.SingleThreadedExperiment', autospec=True) as mock:
        try_command('experiment', 'example/config/experiment.yaml')
        mock.assert_called_once()


def test_cli_multicoreexperiment():
    with patch('triage.cli.MultiCoreExperiment', autospec=True) as mock:
        try_command('experiment', 'example/config/experiment.yaml', '--n-processes', '2')
        mock.assert_called_once()


def test_cli_show_timechop():
    with patch('triage.cli.SingleThreadedExperiment', autospec=True) as exp_mock:
        exp_instance_mock = Mock()
        exp_mock.return_value = exp_instance_mock
        exp_instance_mock.configure_mock(chopper='chopper')
        with patch('triage.cli.ProjectStorage', autospec=True) as ps_mock:
            with patch('triage.cli.visualize_chops', autospec=True) as viz_mock:
                try_command('experiment', 'example/config/experiment.yaml', '--show-timechop')
                ps_mock.assert_called_once()
                viz_mock.assert_called_once()
                assert viz_mock.call_args[1]


def test_cli_audition():
    with patch('triage.cli.AuditionRunner', autospec=True) as mock:
        try_command('audition', '-c', 'example/config/audition.yaml')
        mock.assert_called_once()


def test_cli_crosstabs():
    with patch('triage.cli.run_crosstabs', autospec=True) as mock:
        try_command('crosstabs', 'example/config/postmodeling_crosstabs.yaml')
        mock.assert_called_once()


def test_featuretest():
    with patch('triage.cli.FeatureGenerator', autospec=True) as featuremock:
        with patch('triage.cli.EntityDateTableGenerator', autospec=True) as cohortmock:
            try_command('featuretest', 'example/config/experiment.yaml', '2017-06-06')
            featuremock.assert_called_once()
            cohortmock.assert_called_once()



from triage.component.catwalk import ModelTrainTester, Predictor, ModelTrainer, ModelEvaluator, IndividualImportanceCalculator, ProtectedGroupsGenerator
from triage.component.catwalk.utils import save_experiment_and_get_hash
from triage.component.catwalk.model_trainers import flatten_grid_config
from triage.component.catwalk.storage import (
    ModelStorageEngine,
    MatrixStore,
    MatrixStorageEngine,
)
from tests.utils import (
    get_matrix_store,
    matrix_metadata_creator,
)

from unittest.mock import patch, create_autospec, MagicMock


def test_ModelTrainTester_generate_tasks(db_engine_with_results_schema, project_storage, sample_timechop_splits, sample_grid_config):
    db_engine = db_engine_with_results_schema
    model_storage_engine = ModelStorageEngine(project_storage)
    matrix_storage_engine = MatrixStorageEngine(project_storage)
    sample_matrix_store = get_matrix_store(project_storage)
    experiment_hash = save_experiment_and_get_hash({}, 1234, db_engine)
    # instantiate pipeline objects
    trainer = ModelTrainer(
        experiment_hash=experiment_hash,
        model_storage_engine=model_storage_engine,
        db_engine=db_engine,
    )
    train_tester = ModelTrainTester(
        matrix_storage_engine=matrix_storage_engine,
        model_trainer=trainer,
        model_evaluator=None,
        individual_importance_calculator=None,
        predictor=None,
        subsets=None,
        protected_groups_generator=None,
    )
    with patch.object(matrix_storage_engine, 'get_store', return_value=sample_matrix_store):
        batches = train_tester.generate_task_batches(
            splits=sample_timechop_splits,
            grid_config=sample_grid_config
        )
        assert len(batches) == 3
        # we expect to have a task for each combination of split and classifier
        flattened_tasks = list(task for batch in batches for task in batch.tasks)
        assert len(flattened_tasks) == \
            len(sample_timechop_splits) * len(list(flatten_grid_config(sample_grid_config)))
        # we also expect each task to match the call signature of process_task
        with patch.object(train_tester, 'process_task', autospec=True):
            for task in flattened_tasks:
                train_tester.process_task(**task)


def setup_model_train_tester(project_storage, replace):
    matrix_storage_engine = MatrixStorageEngine(project_storage)
    train_matrix_store = get_matrix_store(
        project_storage,
        metadata=matrix_metadata_creator(matrix_type="train"),
        write_to_db=False
    )
    test_matrix_store = get_matrix_store(
        project_storage,
        metadata=matrix_metadata_creator(matrix_type="test"),
        write_to_db=False
    )
    sample_train_kwargs = {
        'matrix_store': train_matrix_store,
        'class_path': None,
        'parameters': {},
        'model_hash': None,
        'misc_db_parameters': {}
    }
    train_test_task = {
        'train_kwargs': sample_train_kwargs,
        'train_store': train_matrix_store,
        'test_store': test_matrix_store
    }

    predictor = MagicMock(spec_set=create_autospec(Predictor))
    trainer = MagicMock(spec_set=create_autospec(ModelTrainer))
    evaluator = MagicMock(spec_set=create_autospec(ModelEvaluator))
    individual_importance_calculator = MagicMock(spec_set=create_autospec(IndividualImportanceCalculator))
    protected_groups_generator = MagicMock(spec_set=create_autospec(ProtectedGroupsGenerator))
    train_tester = ModelTrainTester(
        matrix_storage_engine=matrix_storage_engine,
        model_trainer=trainer,
        model_evaluator=evaluator,
        individual_importance_calculator=individual_importance_calculator,
        predictor=predictor,
        subsets=[],
        replace=replace,
        protected_groups_generator=protected_groups_generator
    )
    return train_tester, train_test_task


def test_ModelTrainTester_process_task_replace_False_needs_evaluations(project_storage):
    train_tester, train_test_task = setup_model_train_tester(project_storage, replace=False)
    train_tester.model_evaluator.needs_evaluations.return_value = True
    train_tester.process_task(**train_test_task)
    assert train_tester.model_evaluator.needs_evaluations.call_count == 2
    assert train_tester.predictor.predict.call_count == 2
    assert train_tester.model_evaluator.evaluate.call_count == 2
    assert train_tester.protected_groups_generator.as_dataframe.call_count == 2


def test_ModelTrainTester_process_task_replace_False_no_evaluations(project_storage):
    train_tester, train_test_task = setup_model_train_tester(project_storage, replace=False)
    train_tester.model_evaluator.needs_evaluations.return_value = False
    train_tester.process_task(**train_test_task)
    assert train_tester.model_evaluator.needs_evaluations.call_count == 2
    assert train_tester.predictor.predict.call_count == 0
    assert train_tester.model_evaluator.evaluate.call_count == 0
    assert train_tester.protected_groups_generator.as_dataframe.call_count == 0


def test_ModelTrainTester_process_task_replace_True(project_storage):
    train_tester, train_test_task = setup_model_train_tester(project_storage, replace=True)
    train_tester.process_task(**train_test_task)
    assert train_tester.model_evaluator.needs_evaluations.call_count == 0
    assert train_tester.predictor.predict.call_count == 2
    assert train_tester.model_evaluator.evaluate.call_count == 2
    assert train_tester.protected_groups_generator.as_dataframe.call_count == 2


def test_ModelTrainTester_process_task_empty_train(project_storage):
    train_tester, train_test_task = setup_model_train_tester(project_storage, replace=True)
    train_store = MagicMock()
    train_store.empty = True
    train_test_task['train_store'] = train_store
    train_tester.process_task(**train_test_task)

    assert train_tester.model_trainer.process_train_task.call_count == 0
    assert train_tester.model_evaluator.needs_evaluations.call_count == 0
    assert train_tester.predictor.predict.call_count == 0
    assert train_tester.model_evaluator.evaluate.call_count == 0
    assert train_tester.protected_groups_generator.as_dataframe.call_count == 0



from triage.component.audition.selection_rules import BoundSelectionRule
from triage.component.audition.selection_rule_grid import make_selection_rule_grid


def test_selection_rule_grid():
    input_data = [
        {
            "shared_parameters": [
                {"metric": "precision@", "parameter": "100_abs"},
                {"metric": "recall@", "parameter": "100_abs"},
            ],
            "selection_rules": [
                {
                    "name": "most_frequent_best_dist",
                    "dist_from_best_case": [0.1, 0.2, 0.3],
                },
                {"name": "best_current_value"},
            ],
        },
        {
            "shared_parameters": [{"metric1": "precision@", "parameter1": "100_abs"}],
            "selection_rules": [
                {
                    "name": "best_average_two_metrics",
                    "metric2": ["recall@"],
                    "parameter2": ["100_abs"],
                    "metric1_weight": [0.4, 0.5, 0.6],
                }
            ],
        },
    ]

    expected_output = [
        BoundSelectionRule(
            descriptive_name="most_frequent_best_dist_precision@_100_abs_0.1",
            function_name="most_frequent_best_dist",
            args={
                "metric": "precision@",
                "parameter": "100_abs",
                "dist_from_best_case": 0.1,
            },
        ),
        BoundSelectionRule(
            descriptive_name="most_frequent_best_dist_precision@_100_abs_0.2",
            function_name="most_frequent_best_dist",
            args={
                "metric": "precision@",
                "parameter": "100_abs",
                "dist_from_best_case": 0.2,
            },
        ),
        BoundSelectionRule(
            descriptive_name="most_frequent_best_dist_precision@_100_abs_0.3",
            function_name="most_frequent_best_dist",
            args={
                "metric": "precision@",
                "parameter": "100_abs",
                "dist_from_best_case": 0.3,
            },
        ),
        BoundSelectionRule(
            descriptive_name="most_frequent_best_dist_recall@_100_abs_0.1",
            function_name="most_frequent_best_dist",
            args={
                "metric": "recall@",
                "parameter": "100_abs",
                "dist_from_best_case": 0.1,
            },
        ),
        BoundSelectionRule(
            descriptive_name="most_frequent_best_dist_recall@_100_abs_0.2",
            function_name="most_frequent_best_dist",
            args={
                "metric": "recall@",
                "parameter": "100_abs",
                "dist_from_best_case": 0.2,
            },
        ),
        BoundSelectionRule(
            descriptive_name="most_frequent_best_dist_recall@_100_abs_0.3",
            function_name="most_frequent_best_dist",
            args={
                "metric": "recall@",
                "parameter": "100_abs",
                "dist_from_best_case": 0.3,
            },
        ),
        BoundSelectionRule(
            descriptive_name="best_current_value_precision@_100_abs",
            function_name="best_current_value",
            args={"metric": "precision@", "parameter": "100_abs"},
        ),
        BoundSelectionRule(
            descriptive_name="best_current_value_recall@_100_abs",
            function_name="best_current_value",
            args={"metric": "recall@", "parameter": "100_abs"},
        ),
        BoundSelectionRule(
            descriptive_name="best_average_two_metrics_precision@_100_abs_recall@_100_abs_0.4",
            function_name="best_average_two_metrics",
            args={
                "metric1": "precision@",
                "parameter1": "100_abs",
                "metric2": "recall@",
                "parameter2": "100_abs",
                "metric1_weight": 0.4,
            },
        ),
        BoundSelectionRule(
            descriptive_name="best_average_two_metrics_precision@_100_abs_recall@_100_abs_0.5",
            function_name="best_average_two_metrics",
            args={
                "metric1": "precision@",
                "parameter1": "100_abs",
                "metric2": "recall@",
                "parameter2": "100_abs",
                "metric1_weight": 0.5,
            },
        ),
        BoundSelectionRule(
            descriptive_name="best_average_two_metrics_precision@_100_abs_recall@_100_abs_0.6",
            function_name="best_average_two_metrics",
            args={
                "metric1": "precision@",
                "parameter1": "100_abs",
                "metric2": "recall@",
                "parameter2": "100_abs",
                "metric1_weight": 0.6,
            },
        ),
    ]

    # sort both lists so we can compare them without resorting to a hash
    expected_output.sort(key=lambda x: x.descriptive_name)
    grid = sorted(
        make_selection_rule_grid(input_data), key=lambda x: x.descriptive_name
    )
    assert len(grid) == len(expected_output)
    for expected_rule, actual_rule in zip(expected_output, grid):
        assert expected_rule.descriptive_name == actual_rule.descriptive_name



import pandas as pd

from triage.component.catwalk.individual_importance.uniform import uniform_distribution
from tests.utils import rig_engines, get_matrix_store, matrix_metadata_creator
import datetime

from tests.results_tests.factories import (
    ModelFactory,
    FeatureImportanceFactory,
)


def test_uniform_distribution():
    with rig_engines() as (db_engine, project_storage):
        model = ModelFactory()
        feature_importances = [
            FeatureImportanceFactory(model_rel=model, feature="feature_{}".format(i))
            for i in range(0, 10)
        ]
        data_dict = {"entity_id": [1, 1], "as_of_date": ["2016-01-01", "2017-01-01"], "label": [0, 1]}
        for imp in feature_importances:
            data_dict[imp.feature] = [0.5, 0.5]
        metadata = matrix_metadata_creator()
        test_store = get_matrix_store(
            project_storage,
            pd.DataFrame.from_dict(data_dict),
            metadata,
        )
        results = uniform_distribution(
            db_engine,
            model_id=model.model_id,
            as_of_date=datetime.date(2016, 1, 1),
            test_matrix_store=test_store,
            n_ranks=5,
        )

        assert len(results) == 5  # 5 features x 1 entity for this as_of_date
        for result in results:
            assert "entity_id" in result
            assert "feature_name" in result
            assert "score" in result
            assert "feature_value" in result
            assert result["feature_value"] == 0.5
            assert result["score"] >= 0
            assert result["score"] <= 1
            assert isinstance(result["feature_name"], str)
            assert result["entity_id"] in [1, 2]



from unittest.mock import patch

import numpy as np
import testing.postgresql
from sqlalchemy import create_engine

from triage.component.audition.model_group_performance import (
    ModelGroupPerformancePlotter,
)

from .utils import create_sample_distance_table


def test_ModelGroupPerformancePlotter_generate_plot_data():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        distance_table, model_groups = create_sample_distance_table(engine)
        plotter = ModelGroupPerformancePlotter(distance_table)
        df = plotter.generate_plot_data(
            metric="precision@",
            parameter="100_abs",
            model_group_ids=[1, 2],
            train_end_times=["2014-01-01", "2015-01-01"],
        )
        assert sorted(df["model_type"].unique()) == [
            "best case",
            "mySpikeClassifier",
            "myStableClassifier",
        ]
        for value in df[df["model_group_id"] == 1]["raw_value"].values:
            assert np.isclose(value, 0.5)


def test_ModelGroupPerformancePlotter_plot_all():
    with patch(
        "triage.component.audition.model_group_performance.plot_cats"
    ) as plot_patch:
        with testing.postgresql.Postgresql() as postgresql:
            engine = create_engine(postgresql.url())
            distance_table, model_groups = create_sample_distance_table(engine)
            plotter = ModelGroupPerformancePlotter(distance_table)
            plotter.plot_all(
                [{"metric": "precision@", "parameter": "100_abs"}],
                model_group_ids=[1, 2],
                train_end_times=["2014-01-01", "2015-01-01"],
            )
        assert plot_patch.called
        args, kwargs = plot_patch.call_args
        assert "raw_value" in kwargs["frame"]
        assert "train_end_time" in kwargs["frame"]
        assert kwargs["x_col"] == "train_end_time"
        assert kwargs["y_col"] == "raw_value"



import testing.postgresql
import datetime
from copy import copy

from sqlalchemy import create_engine
from triage.component.catwalk.db import ensure_db

from triage.component.catwalk.model_grouping import ModelGrouper
from .utils import sample_metadata


def test_model_grouping_default_config(sample_metadata):
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        ensure_db(engine)
        model_grouper = ModelGrouper()
        # get the basic first model group with our default matrix
        assert (
            model_grouper.get_model_group_id(
                "module.Classifier", {"param1": "val1"}, sample_metadata, engine
            )
            == 1
        )

        # the end time is not by default a model group key so changing it
        # should still get us the same group
        metadata_new_end_time = copy(sample_metadata)
        metadata_new_end_time["end_time"] = datetime.date(2017, 3, 20)
        assert (
            model_grouper.get_model_group_id(
                "module.Classifier", {"param1": "val1"}, metadata_new_end_time, engine
            )
            == 1
        )

        # max_training_history is a default key,
        # so it should trigger a new group
        metadata_train_history = copy(sample_metadata)
        metadata_train_history["max_training_history"] = "3y"
        assert (
            model_grouper.get_model_group_id(
                "module.Classifier", {"param1": "val1"}, metadata_train_history, engine
            )
            == 2
        )

        # classifier is of course a default key as well
        assert (
            model_grouper.get_model_group_id(
                "module.OtherClassifier", {"param1": "val1"}, sample_metadata, engine
            )
            == 3
        )


def test_model_grouping_custom_config(sample_metadata):
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        ensure_db(engine)
        model_grouper = ModelGrouper(
            model_group_keys=["feature_names", "as_of_date_frequency"]
        )
        # get the basic first model group with our default matrix
        assert (
            model_grouper.get_model_group_id(
                "module.Classifier", {"param1": "val1"}, sample_metadata, engine
            )
            == 1
        )

        # classifier is now not a key, so changing it should not get a new id
        assert (
            model_grouper.get_model_group_id(
                "module.OtherClassifier", {"param1": "val1"}, sample_metadata, engine
            )
            == 1
        )

        # as_of_date_frequency is a key,
        # so it should trigger a new group
        metadata_frequency = copy(sample_metadata)
        metadata_frequency["as_of_date_frequency"] = "2w"
        assert (
            model_grouper.get_model_group_id(
                "module.Classifier", {"param1": "val1"}, metadata_frequency, engine
            )
            == 2
        )

        # testing feature names may seem redundant but it is on a separate
        # code path so make sure its logic works
        metadata_features = copy(sample_metadata)
        metadata_features["feature_names"] = ["ft1", "ft3"]
        assert (
            model_grouper.get_model_group_id(
                "module.Classifier", {"param1": "val1"}, metadata_features, engine
            )
            == 3
        )



from triage.component.architect.features import FeatureDictionaryCreator
import testing.postgresql
from sqlalchemy import create_engine


def test_feature_dictionary_creator():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        engine.execute("create schema features")
        engine.execute(
            """
            create table features.prefix1_entity_id (
                entity_id int,
                as_of_date date,
                feature_one float,
                feature_two float
            )
        """
        )
        engine.execute(
            """
            create table features.prefix1_zipcode (
                zipcode text,
                as_of_date date,
                feature_three float,
                feature_four float
            )
        """
        )
        engine.execute(
            """
            create table features.prefix1_aggregation (
                entity_id int,
                as_of_date date,
                zipcode text,
                feature_one float,
                feature_two float,
                feature_three float,
                feature_four float
            )
        """
        )
        engine.execute(
            """
            create table features.prefix1_aggregation_imputed (
                entity_id int,
                as_of_date date,
                zipcode text,
                feature_one float,
                feature_two float,
                feature_three float,
                feature_three_imp int,
                feature_four float
            )
        """
        )
        engine.execute(
            """
            create table features.random_other_table (
                another_column float
            )
        """
        )

        creator = FeatureDictionaryCreator(
            features_schema_name="features", db_engine=engine
        )
        feature_dictionary = creator.feature_dictionary(
            feature_table_names=[
                "prefix1_entity_id",
                "prefix1_zip_code",
                "prefix1_aggregation",
                "prefix1_aggregation_imputed",
            ],
            index_column_lookup={
                "prefix1_aggregation_imputed": ["entity_id", "zipcode", "as_of_date"]
            },
        )
        assert feature_dictionary == {
            "prefix1_aggregation_imputed": [
                "feature_one",
                "feature_two",
                "feature_three",
                "feature_three_imp",
                "feature_four",
            ]
        }



# -*- coding: utf-8 -*-
from .collate import available_imputations, Aggregation, Aggregate, Compare, Categorical
from .from_obj import FromObj
from .spacetime import SpacetimeAggregation

__all__ = [
    "available_imputations",
    "Aggregation",
    "Aggregate",
    "FromObj",
    "Compare",
    "Categorical",
    "SpacetimeAggregation",
]
__author__ = """DSaPP Researchers"""
__email__ = "datascifellows@gmail.com"



from triage.component.postmodeling.contrast.model_group_evaluator import ModelGroupEvaluator
import pandas as pd
import pytest
from tests.utils import assert_plot_figures_added


@pytest.fixture(scope="module")
def model_group_evaluator(finished_experiment):
    return ModelGroupEvaluator((1,1), finished_experiment.db_engine)


def test_ModelGroupEvaluator_metadata(model_group_evaluator):
    assert isinstance(model_group_evaluator.metadata, list)
    assert len(model_group_evaluator.metadata) == 8 # 8 model groups expected from basic experiment
    for row in model_group_evaluator.metadata:
        assert isinstance(row, dict)
        

def test_ModelGroupEvaluator_model_type(model_group_evaluator):
    assert model_group_evaluator.model_type[0] == 'sklearn.tree.DecisionTreeClassifier'


def test_ModelGroupEvaluator_predictions(model_group_evaluator):
    assert isinstance(model_group_evaluator.predictions, pd.DataFrame)


def test_ModelGroupEvaluator_feature_importances(model_group_evaluator):
    assert isinstance(model_group_evaluator.feature_importances, pd.DataFrame)


def test_ModelGroupEvaluator_metrics(model_group_evaluator):
    assert isinstance(model_group_evaluator.metrics, pd.DataFrame)


def test_ModelGroupEvaluator_feature_groups(model_group_evaluator):
    assert isinstance(model_group_evaluator.feature_groups, pd.DataFrame)


def test_ModelGroupEvaluator_same_time_models(model_group_evaluator):
    assert isinstance(model_group_evaluator.same_time_models, pd.DataFrame)


def test_ModelGroupEvaluator_plot_prec_across_time(model_group_evaluator):
    with assert_plot_figures_added():
        model_group_evaluator.plot_prec_across_time()


def test_ModelGroupEvaluator_feature_loi_loo(model_group_evaluator):
    with pytest.raises(IndexError):
        model_group_evaluator.feature_loi_loo()


def test_ModelGroupEvaluator_plot_ranked_correlation_preds(model_group_evaluator):
    with assert_plot_figures_added():
        model_group_evaluator.plot_ranked_correlation_preds(param_type='rank_abs', param=10, top_n_features=10)


def test_ModelGroupEvaluator_plot_ranked_correlation_features(model_group_evaluator):
    with assert_plot_figures_added():
        model_group_evaluator.plot_ranked_correlation_features(param_type='rank_abs', param=10, top_n_features=10)

def test_ModelGroupEvaluator_plot_jaccard_preds(model_group_evaluator):
    with assert_plot_figures_added():
        model_group_evaluator.plot_jaccard_preds(param_type='rank_abs', param=10)


def test_ModelGroupEvaluator_plot_jaccard_features(model_group_evaluator):
    with assert_plot_figures_added():
        model_group_evaluator.plot_jaccard_features()


def test_ModelGroupEvaluator_plot_preds_comparison(model_group_evaluator):
    with assert_plot_figures_added():
        model_group_evaluator.plot_preds_comparison(param_type='rank_abs', param=10)



"""evaluation_randomness

Revision ID: 97cf99b7348f
Revises: cfd5c3386014
Create Date: 2019-04-11 13:21:09.907895

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '97cf99b7348f'
down_revision = 'cfd5c3386014'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('evaluations', sa.Column('best_value', sa.Numeric(), nullable=True), schema='train_results')
    op.add_column('evaluations', sa.Column('worst_value', sa.Numeric(), nullable=True), schema='train_results')
    op.add_column('evaluations', sa.Column('stochastic_value', sa.Numeric(), nullable=True), schema='train_results')
    op.add_column('evaluations', sa.Column('num_sort_trials', sa.Integer(), nullable=True), schema='train_results')
    op.add_column('evaluations', sa.Column('standard_deviation', sa.Numeric(), nullable=True), schema='train_results')
    op.add_column('evaluations', sa.Column('best_value', sa.Numeric(), nullable=True), schema='test_results')
    op.add_column('evaluations', sa.Column('worst_value', sa.Numeric(), nullable=True), schema='test_results')
    op.add_column('evaluations', sa.Column('stochastic_value', sa.Numeric(), nullable=True), schema='test_results')
    op.add_column('evaluations', sa.Column('num_sort_trials', sa.Integer(), nullable=True), schema='test_results')
    op.add_column('evaluations', sa.Column('standard_deviation', sa.Numeric(), nullable=True), schema='test_results')
    op.execute("update train_results.evaluations set stochastic_value = value, num_sort_trials = 1")
    op.execute("update test_results.evaluations set stochastic_value = value, num_sort_trials = 1")
    op.drop_column('evaluations', 'value', schema='train_results')
    op.drop_column('evaluations', 'value', schema='test_results')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('evaluations', sa.Column('value', sa.Numeric(), nullable=True), schema='train_results')
    op.add_column('evaluations', sa.Column('value', sa.Numeric(), nullable=True), schema='test_results')
    op.execute("update train_results.evaluations set value = stochastic_value")
    op.execute("update test_results.evaluations set value = stochastic_value")
    op.drop_column('evaluations', 'worst_value', schema='test_results')
    op.drop_column('evaluations', 'stochastic_value', schema='test_results')
    op.drop_column('evaluations', 'best_value', schema='test_results')
    op.drop_column('evaluations', 'num_sort_trials', schema='test_results')
    op.drop_column('evaluations', 'standard_deviation', schema='test_results')
    op.drop_column('evaluations', 'worst_value', schema='train_results')
    op.drop_column('evaluations', 'stochastic_value', schema='train_results')
    op.drop_column('evaluations', 'best_value', schema='train_results')
    op.drop_column('evaluations', 'num_sort_trials', schema='train_results')
    op.drop_column('evaluations', 'standard_deviation', schema='train_results')
    # ### end Alembic commands ###



import inspect
import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

from numpy import exp, log, average

from .metric_directionality import greater_is_better, best_in_series, idxbest


def random_model_group(df, train_end_time, n=1):
    """Pick a random model group (as a baseline)

    Arguments:
        train_end_time (Timestamp): current train end time
        df (pandas.DataFrame): dataframe containing the columns
                model_group_id,
                train_end_time,
                metric,
                parameter,
                raw_value,
                below_best
        n (int): the number of model group ids to return
    Returns: (int) the model group id to select, with highest current raw metric value
    """
    return df["model_group_id"].drop_duplicates().sample(frac=1).tolist()[:n]


def _mg_best_avg_by(df, value_col, metric, n=1):
    """Best model group in dataframe by average of some column

    Args:
        df (pandas.DataFrame)
        value_col (str): The column which contains the value to be averaged
        metric (str): the name of the column
        n (int): the number of model group ids to return
    """
    if n == 1:
        return [
            getattr(
                df.groupby(["model_group_id"])[value_col].mean().sample(frac=1),
                idxbest(metric),
            )()
        ]
    else:
        if greater_is_better(metric):
            return (
                df.groupby(["model_group_id"])[value_col]
                .mean()
                .nlargest(n)
                .index.tolist()
            )
        else:
            return (
                df.groupby(["model_group_id"])[value_col]
                .mean()
                .nsmallest(n)
                .index.tolist()
            )


def best_current_value(df, train_end_time, metric, parameter, n=1):
    """Pick the model group with the best current metric value

    Arguments:
        metric (string): model evaluation metric, such as 'precision@'
        parameter (string): model evaluation metric parameter,
            such as '300_abs'
        train_end_time (Timestamp): current train end time
        df (pandas.DataFrame): dataframe containing the columns:
                model_group_id,
                train_end_time,
                metric,
                parameter,
                raw_value,
                dist_from_best_case
        n (int): the number of model group ids to return
    Returns: (int) the model group id to select, with highest current raw metric value
    """
    curr_df = df.loc[
        (df["train_end_time"] == train_end_time)
        & (df["metric"] == metric)
        & (df["parameter"] == parameter)
    ]
    # sample(frac=1) to shuffle rows so we don't accidentally introduce bias in breaking ties
    best_raw_value = getattr(curr_df["raw_value"], best_in_series(metric))()
    if n <= 1:
        return (
            curr_df.loc[curr_df["raw_value"] == best_raw_value, "model_group_id"]
            .sample(frac=1)
            .tolist()
        )
    else:
        if greater_is_better(metric):
            result = curr_df.nlargest(n, "raw_value")["model_group_id"].tolist()
            return result
        else:
            result = curr_df.nsmallest(n, "raw_value")["model_group_id"].tolist()
            return result


def best_average_value(df, train_end_time, metric, parameter, n=1):
    """Pick the model with the highest average metric value so far

    Arguments:
        metric (string): model evaluation metric, such as 'precision@'
        parameter (string): model evaluation metric parameter,
            such as '300_abs'
        train_end_time (Timestamp): current train end time
        df (pandas.DataFrame): dataframe containing the columns
                model_group_id,
                train_end_time,
                metric,
                parameter,
                raw_value,
                dist_from_best_case
        n (int): the number of model group ids to return
    Returns: (int) the model group id to select, with highest mean raw metric value
    """
    met_df = df.loc[(df["metric"] == metric) & (df["parameter"] == parameter)]
    return _mg_best_avg_by(met_df, "raw_value", metric, n)


def lowest_metric_variance(df, train_end_time, metric, parameter, n=1):
    """Pick the model with the lowest metric variance so far

    Arguments:
        metric (string): model evaluation metric, such as 'precision@'
        parameter (string): model evaluation metric parameter,
            such as '300_abs'
        train_end_time (Timestamp): current train end time
        df (pandas.DataFrame): dataframe containing the columns
                model_group_id,
                train_end_time,
                metric,
                parameter,
                raw_value,
                below_best
        n (int): the number of model group ids to return
    Returns: (int) the model group id to select, with highest mean raw metric value
    """

    met_df = (
        df.loc[(df["metric"] == metric) & (df["parameter"] == parameter)]
        .groupby(["model_group_id"])["raw_value"]
        .std()
    )

    if met_df.isnull().sum() == met_df.shape[0]:
        # variance will be undefined in first time window since we only have one obseravtion
        # per model group
        logger.debug(
            "Null metric variances for %s %s at %s; picking at random",
            metric,
            parameter,
            train_end_time,
        )
        return df["model_group_id"].drop_duplicates().sample(n=n).tolist()
    elif met_df.isnull().sum() > 0:
        # the variances should be all null or no nulls, a mix shouldn't be possible
        # since we should have the same number of observations for every model group
        raise ValueError(
            "Mix of null and non-null metric variances for or {} {} at {}".format(
                metric, parameter, train_end_time
            )
        )
    if n == 1:
        # sample(frac=1) to shuffle rows so we don't accidentally introduce bias in breaking ties
        return [met_df.sample(frac=1).idxmin()]
    else:
        return met_df.nsmallest(n).index.tolist()


def most_frequent_best_dist(
    df, train_end_time, metric, parameter, dist_from_best_case, n=1
):
    """Pick the model that is most frequently within `dist_from_best_case` from the
    best-performing model group across test sets so far

    Arguments:
        dist_from_best_case (float): distance from the best performing model
        metric (string): model evaluation metric, such as 'precision@'
        parameter (string): model evaluation metric parameter,
            such as '300_abs'
        train_end_time (Timestamp): current train end time
        df (pandas.DataFrame): dataframe containing the columns
                model_group_id,
                train_end_time,
                metric,
                parameter,
                raw_value,
                below_best
        n (int): the number of model group ids to return
    Returns: (int) the model group id to select, with highest mean raw metric value
    """

    met_df = df.loc[(df["metric"] == metric) & (df["parameter"] == parameter)]
    met_df["within_dist"] = (df["dist_from_best_case"] <= dist_from_best_case).astype(
        "int"
    )
    if n == 1:
        # sample(frac=1) to shuffle rows so we don't accidentally introduce bias in breaking ties
        return [
            met_df.groupby(["model_group_id"])["within_dist"]
            .mean()
            .sample(frac=1)
            .idxmax()
        ]
    else:
        return (
            met_df.groupby(["model_group_id"])["within_dist"]
            .mean()
            .nlargest(n)
            .index.tolist()
        )


def best_average_two_metrics(
    df,
    train_end_time,
    metric1,
    parameter1,
    metric2,
    parameter2,
    metric1_weight=0.5,
    n=1,
):
    """Pick the model with the highest average combined value to date
    of two metrics weighted together using `metric1_weight`

    Arguments:
        metric1_weight (float): relative weight of metric1, between 0 and 1
        metric1 (string): model evaluation metric, such as 'precision@'
        parameter1 (string): model evaluation metric parameter,
            such as '300_abs'
        metric2 (string): model evaluation metric, such as 'precision@'
        parameter2 (string): model evaluation metric parameter,
            such as '300_abs'
        train_end_time (Timestamp): current train end time
        df (pandas.DataFrame): dataframe containing the columns
                model_group_id,
                train_end_time,
                metric,
                parameter,
                raw_value,
                below_best
        n (int): the number of model group ids to return
    Returns: (int) the model group id to select, with highest mean raw metric value
    """

    if metric1_weight < 0 or metric1_weight > 1:
        raise ValueError("Metric weight must be between 0 and 1")

    metric1_dir = greater_is_better(metric1)
    metric2_dir = greater_is_better(metric2)
    if metric1_dir != metric2_dir:
        raise ValueError("Metric directionalities must be the same")

    met_df = df.loc[
        ((df["metric"] == metric1) & (df["parameter"] == parameter1))
        | ((df["metric"] == metric2) & (df["parameter"] == parameter2))
    ]

    met_df.loc[
        (met_df["metric"] == metric1) & (met_df["parameter"] == parameter1),
        "weighted_raw",
    ] = (
        met_df.loc[
            (met_df["metric"] == metric1) & (met_df["parameter"] == parameter1),
            "raw_value",
        ]
        * metric1_weight
    )

    met_df.loc[
        (met_df["metric"] == metric2) & (met_df["parameter"] == parameter2),
        "weighted_raw",
    ] = met_df.loc[
        (met_df["metric"] == metric2) & (met_df["parameter"] == parameter2), "raw_value"
    ] * (
        1.0 - metric1_weight
    )

    met_df_wt = met_df.groupby(
        ["model_group_id", "train_end_time"], as_index=False
    ).sum()

    # sample(frac=1) to shuffle rows so we don't accidentally introduce bias in breaking ties
    return _mg_best_avg_by(met_df_wt, "weighted_raw", metric1, n)


def best_avg_var_penalized(df, train_end_time, metric, parameter, stdev_penalty, n=1):
    """Pick the model with the highest
     average metric value so far, placing less weight in older
     results. You need to specify two parameters: the shape of how the
     weight affects points (decay_type, linear or exponential) and the relative
     weight of the most recent point (curr_weight).

    Arguments:
        stdev_penalty (float): penalty for instability
        metric (string): model evaluation metric, such as 'precision@'
        parameter (string): model evaluation metric parameter,
            such as '300_abs'
        train_end_time (Timestamp): current train end time
        df (pandas.DataFrame): dataframe containing the columns
                model_group_id,
                train_end_time,
                metric,
                parameter,
                raw_value,
                below_best
        n (int): the number of model group ids to return
    Returns: (int) the model group id to select, with highest mean raw metric value
    """

    # for metrics where smaller values are better, the penalty for instability should
    # add to the mean, so introduce a factor of -1
    stdev_penalty = stdev_penalty if greater_is_better(metric) else -1.0 * stdev_penalty

    met_df = df.loc[(df["metric"] == metric) & (df["parameter"] == parameter)]
    met_df_grp = met_df.groupby(["model_group_id"]).aggregate(
        {"raw_value": ["mean", "std"]}
    )
    met_df_grp.columns = met_df_grp.columns.droplevel(0)
    met_df_grp.columns = ["raw_avg", "raw_stdev"]
    if met_df_grp["raw_stdev"].isnull().sum() == met_df_grp.shape[0]:
        # variance will be undefined in first time window since we only have one obseravtion
        # per model group
        logger.debug(
            "Null metric variances for %s %s at %s; just using mean",
            metric,
            parameter,
            train_end_time,
        )
        return [getattr(met_df_grp["raw_avg"].sample(frac=1), idxbest(metric))()]
    elif met_df_grp["raw_stdev"].isnull().sum() > 0:
        # the variances should be all null or no nulls, a mix shouldn't be possible
        # since we should have the same number of observations for every model group
        raise ValueError(
            "Mix of null and non-null metric variances for or {} {} at {}".format(
                metric, parameter, train_end_time
            )
        )

    min_stdev = met_df_grp["raw_stdev"].min()
    met_df_grp["penalized_avg"] = met_df_grp["raw_avg"] - stdev_penalty * (
        met_df_grp["raw_stdev"] - min_stdev
    )

    if n == 1:
        # sample(frac=1) to shuffle rows so we don't accidentally introduce bias in breaking ties
        return [getattr(met_df_grp["penalized_avg"].sample(frac=1), idxbest(metric))()]
    else:
        if greater_is_better(metric):
            return met_df_grp["penalized_avg"].nlargest(n).index.tolist()
        else:
            return met_df_grp["penalized_avg"].nsmallest(n).index.tolist()


def best_avg_recency_weight(
    df, train_end_time, metric, parameter, curr_weight, decay_type, n=1
):
    """Pick the model with the highest average metric value so far, penalized
    for relative variance as:
        avg_value - (stdev_penalty) * (stdev - min_stdev)
    where min_stdev is the minimum standard deviation of the metric across all
    model groups

    Arguments:
        decay_type (string): either 'linear' or 'exponential'; the shape of
            how the weights fall off between the current and first point
        curr_weight (float): amount of weight to put on the most recent point,
            relative to the first point (e.g., a value of 5.0 would mean the
            current data is weighted 5 times as much as the first one)
        metric (string): model evaluation metric, such as 'precision@'
        parameter (string): model evaluation metric parameter,
            such as '300_abs'
        train_end_time (Timestamp): current train end time
        df (pandas.DataFrame): dataframe containing the columns
                model_group_id,
                train_end_time,
                metric,
                parameter,
                raw_value,
                below_best
        n (int): the number of model group ids to return
    Returns: (int) the model group id to select, with highest mean raw metric value
    """

    # curr_weight is amount of weight to put on current point, relative to the first point
    # (e.g., if the first point has a weight of 1.0)
    # decay type is linear or exponetial

    first_date = df["train_end_time"].min()
    df["days_out"] = (df["train_end_time"] - first_date).apply(lambda x: float(x.days))
    tmax = df["days_out"].max()

    if tmax == 0:
        # only one date (must be on first time point), so everything gets a weight of 1
        df["weight"] = 1.0
    elif decay_type == "linear":
        # weight = (curr_weight - 1.0) * (t/tmax) + 1.0
        df["weight"] = (curr_weight - 1.0) * (df["days_out"] / tmax) + 1.0
    elif decay_type == "exponential":
        # weight = exp(ln(curr_weight)*t/tmax)
        df["weight"] = exp(log(curr_weight) * df["days_out"] / tmax)
    else:
        raise ValueError("Must specify linear or exponential decay type")

    def wm(x):
        return average(x, weights=df.loc[x.index, "weight"])

    met_df = df.loc[(df["metric"] == metric) & (df["parameter"] == parameter)]
    if n == 1:
        # sample(frac=1) to shuffle rows so we don't accidentally introduce bias in breaking ties
        result = getattr(
            met_df.groupby(["model_group_id"])
            .aggregate({"raw_value": wm})
            .sample(frac=1),
            idxbest(metric),
        )()
        return result.tolist()
    else:
        met_df_grp = met_df.groupby(["model_group_id"]).aggregate({"raw_value": wm})
        if greater_is_better(metric):

            return met_df_grp["raw_value"].nlargest(n).index.tolist()
        else:
            return met_df_grp["raw_value"].nsmallest(n).index.tolist()


SELECTION_RULES = {
    "random_model_group": random_model_group,
    "best_current_value": best_current_value,
    "best_average_value": best_average_value,
    "lowest_metric_variance": lowest_metric_variance,
    "most_frequent_best_dist": most_frequent_best_dist,
    "best_average_two_metrics": best_average_two_metrics,
    "best_avg_var_penalized": best_avg_var_penalized,
    "best_avg_recency_weight": best_avg_recency_weight,
}


class BoundSelectionRule:
    """A selection rule bound with a set of arguments

    Args:
        args (dict): A set of keyword arguments, that should be sufficient
            to call the function when a dataframe and train_end_time is added
        function_name (string, optional): The name of a function in SELECTION_RULES
        descriptive_name (string, optional): A descriptive name, used in charts
            If none is given it will be automatically constructed
        function (function, optional): A function
    """

    def __init__(self, args, function_name=None, descriptive_name=None, function=None):
        if not function_name and not function:
            raise ValueError("Need either function_name or function")

        if not descriptive_name and not function_name:
            raise ValueError("Need either descriptive_name or function_name")

        self.args = args
        self.function_name = function_name
        self._function = function
        self._descriptive_name = descriptive_name

    @property
    def function(self):
        if not self._function:
            self._function = SELECTION_RULES[self.function_name]
        return self._function

    @property
    def descriptive_name(self):
        if not self._descriptive_name:
            self._descriptive_name = self._build_descriptive_name()

        return self._descriptive_name

    def __str__(self):
        return self.descriptive_name

    def _build_descriptive_name(self):
        """Build a descriptive name for the bound selection rule

        Constructed using the function name and arguments.
        """
        argspec = inspect.getfullargspec(self.function)
        args = [arg for arg in argspec.args if arg not in ["df", "train_end_time", "n"]]
        return "_".join([self.function_name] + [str(self.args[key]) for key in args])

    def pick(self, dataframe, train_end_time):
        """Run the selection rule for a given time on a dataframe

        Args:
            dataframe (pandas.DataFrame)
            train_end_time (timestamp) Current train end time

        Returns: (int) a model group id
        """
        return self.function(dataframe, train_end_time, **(self.args))



import pandas as pd

def get_aequitas_results(engine, parameter, schema="test_results", table="aequitas", model_id=None,  subset_hash="", tie_breaker="worst"):
    ''' This function returns the current contents of the aequitas table.

        Args:
            - engine: SQLAlchemy engine conected to database
            - parameter: A string that indicates any parameters for the metric (ex. `100_abs` indicates top-100 entities)
            - schema: Databse schema to find table within
            - table: Databse table to select data from
            - model_id: A model_id, to query only for results of that model
            - subset_hash: Identifies the subset for the evaluation
            - tie_breaker: Indicates how ties are broken

        Returns: A DataFrame, corresponding to schema.table
    '''

    query = f"""SELECT * FROM {schema}.{table} 
                 WHERE parameter = '{parameter}' 
                   AND subset_hash = '{subset_hash}'
                   AND tie_breaker = '{tie_breaker}'
                   """
    if model_id:
        query += f" AND model_id = {model_id}"
    return pd.read_sql(query, con=engine)


import pandas as pd
import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

from scipy import stats
import yaml


class CrosstabsConfigLoader:
    def __init__(self, config_path=None, config=None):
        self.output = {}
        self.models_list_query = None
        self.as_of_dates_query = None
        self.models_dates_join_query = None
        self.features_query = None
        self.predictions_query = None

        if config:
            config_to_load = config
        elif config_path:
            with open(config_path, "r") as stream:
                config_to_load = yaml.full_load(stream)
        else:
            raise ValueError("Either config_path or config are needed")
        self.__dict__.update(config_to_load)


# Later to be run against feature matrices of high- and low-risk entities.
# Each function receives two arguments: A dataframe of high-risk entities'
# features (as columns), and a dataframe of low-risk entities' features.
# The functions return either a Series, where the row index are the names of
# the feature columns, or a dataframe, again with the rows indexed by feature
# names.

high_risk_n = lambda hr, lr: pd.Series(index=hr.columns, data=hr.shape[0])
low_risk_n = lambda hr, lr: pd.Series(index=lr.columns, data=lr.shape[0])
high_risk_mean = lambda hr, lr: hr.mean(axis=0)
low_risk_mean = lambda hr, lr: lr.mean(axis=0)
high_risk_std = lambda hr, lr: hr.std(axis=0)
low_risk_std = lambda hr, lr: lr.std(axis=0)
hr_lr_ratio = lambda hr, lr: hr.mean(axis=0) / lr.mean(axis=0)


def hr_lr_ttest(hr, lr):
    """ Returns the t-test (T statistic and p value), comparing the features for
    high- and low-risk entities. """
    res = stats.ttest_ind(hr.to_numpy(), lr.to_numpy(), axis=0, nan_policy="omit", equal_var=False)

    r0 = pd.Series(res[0], index=hr.columns)
    r1 = pd.Series(res[1], index=hr.columns)

    return pd.DataFrame({"ttest_T": r0, "ttest_p": r1})


# the crosstab functions are passed as a list of tuples;
# the first tuple element gives the name (or names) of the
# metric that the function calculates
crosstab_functions = [
    ("count_predicted_positive", high_risk_n),
    ("count_predicted_negative", low_risk_n),
    ("mean_predicted_positive", high_risk_mean),
    ("mean_predicted_negative", low_risk_mean),
    ("std_predicted_positive", high_risk_std),
    ("std_predicted_negative", low_risk_std),
    ("ratio_predicted_positive_over_predicted_negative", hr_lr_ratio),
    (["ttest_T", "ttest_p"], hr_lr_ttest),
]


def list_to_str(list_sql):
    str = "(" + ",".join(list_sql) + ")"
    return str


def populate_crosstabs_table(
    model_id,
    as_of_date,
    df,
    thresholds=None,
    crosstab_functions=crosstab_functions,
    push_to_db=True,
    return_df=False,
    engine=None,
    schema="results",
    table="crosstabs",
):
    """ This function populates the results.crosstabs table.

        Args:
            entity_id_list (list): A list of entity_ids. Crosstabs will
                                   be calculated using only predictions and
                                   features of these entities. This list is
                                   being passed to get_predictions_query and
                                   get_features_query.
            crosstab_functions (dict): A dictionary of function names and
                                   functions. The functions receive the feature
                                   dataframes for entities that were predicted
                                   postive or negative, respectively. Each function
                                   returns a Series or DataFrame with rows now
                                   indexed by the feature names, and columns
                                   corresponding to the statistics that the
                                   function calculated.
            thresholds (dict): A dictionary that maps column names to lists of
                               threshold values. The (name, threshold) pairs will
                               be applied to the dataframes with predictions to
                               split low- and high-risk entities.
            push_to_db (bool): If True, the crosstab table is being populated.
            return_df (bool): If True, a dataframe corresponding to the crosstabe
                              table will be returned.
            engine (SQLAlchemy engine): DB connection
            schema (str): Name of the schema to store results; defaults to 'results'
            table (str): Name of table to store results; defaults to 'crosstabs'

            Returns: A DataFrame, corresponding to results.crosstabs
    """
    print("\n\n****************\nRUNNING populate_crosstabs_table for model_id ", model_id, "and as_of_date ", as_of_date)
    if len(df) == 0:
        raise ValueError("No data could be fetched.")

    dfs = []

    # this will be useful later - the non-feature columns we'll grab
    model_cols = ["model_id", "as_of_date", "entity_id", "score", "rank_abs", "rank_pct", "label_value"]

    # drop the all-null columns
    null_cols = df.columns[df.isnull().sum() == df.shape[0]]
    null_cols = [c for c in null_cols if c not in model_cols]

    if len(null_cols) > 0:
        logger.warning("Dropping the all-null columns: %s" % str(null_cols))
        df = df.drop(null_cols, 1)

    # if the dataframe isn't dummified, do it
    if object in df.dtypes.values or pd.Categorical.dtype in df.dtypes.values:
        to_dummify = [c for c in df.select_dtypes(include=["category", object]) if c not in model_cols]
        print("Dummifying the data")
        df = pd.get_dummies(df, columns=to_dummify, dummy_na=True)

    feat_cols = [c for c in df.columns if c not in model_cols]
    print("Iterating over thresholds to generate results...")
    for thres_type, thress in thresholds.items():
        for thres in thress:

            results = pd.DataFrame(index=feat_cols)
            print("split dataframe in high risk and lowriks")
            # split dataframe into high/low risk
            df_pred_pos = df.loc[df[thres_type] <= thres, feat_cols]
            df_pred_neg = df.loc[df[thres_type] > thres, feat_cols]
            print("len of hr and lr", len(df_pred_pos), len(df_pred_neg))
            for name, func in crosstab_functions:
                print(name)
                this_result = pd.DataFrame(func(df_pred_pos, df_pred_neg))
                if name in ["ttest_T", "ttest_p"]:
                    print("this_result:", this_result.shape)
                    print("Results:", results.shape)
                if not type(name) in [list, tuple]:
                    name = [name]
                # the metric name is coming from the crosstab_functions tuples
                this_result.columns = name
                results = results.join(this_result, how="outer")

            results = results.stack()
            results.index.names = ["feature_column", "metric"]
            results.name = "value"
            results = results.to_frame()

            results["model_id"] = model_id
            results["as_of_date"] = as_of_date
            results["threshold_unit"] = thres_type[-3:]
            results["threshold_value"] = thres

            dfs.append(results)

    df = pd.concat(dfs)
    if push_to_db:
        print("Pushing results to database...")
        df.reset_index().set_index(["model_id", "as_of_date", "metric"]).to_sql(schema=schema, name=table, con=engine, if_exists="append")
    if return_df:
        return df


def run_crosstabs(db_engine, crosstabs_config):
    crosstabs_query = """
        with models_list_query as (
        {models_list_query}
        ), as_of_dates_query as (
        {as_of_dates_query}
        ),models_dates_join_query as (
        {models_dates_join_query}
        ),features_query as (
        {features_query}
        ), predictions_query as (
        {predictions_query}
        )
        select * from predictions_query
        left join features_query f using (model_id,entity_id, as_of_date)
        """.format(
        models_list_query=crosstabs_config.models_list_query,
        as_of_dates_query=crosstabs_config.as_of_dates_query,
        models_dates_join_query=crosstabs_config.models_dates_join_query,
        features_query=crosstabs_config.features_query,
        predictions_query=crosstabs_config.predictions_query,
    )
    if len(crosstabs_config.entity_id_list) > 0:
        crosstabs_query += " where entity_id=ANY('{%s}') " % ", ".join(map(str, crosstabs_config.entity_id_list))
    crosstabs_query += "  order by model_id, as_of_date, rank_abs asc;"
    df = pd.read_sql(crosstabs_query, db_engine)
    if len(df) == 0:
        raise ValueError("No data could be fetched.")
    groupby_obj = df.groupby(["model_id", "as_of_date"])
    for group, values in groupby_obj:
        df_modelid_asofdate = groupby_obj.get_group(group)
        res = populate_crosstabs_table(
            model_id=group[0],
            as_of_date=group[1],
            df=df_modelid_asofdate,
            thresholds=crosstabs_config.thresholds,
            crosstab_functions=crosstab_functions,
            push_to_db=True,
            return_df=False,
            engine=db_engine,
            schema=crosstabs_config.output["schema"],
            table=crosstabs_config.output["table"],
        )



"""empty message

Revision ID: a20104116533
Revises: 8cef808549dd
Create Date: 2020-06-11 16:32:41.319128

"""
import os
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'a20104116533'
down_revision = '8cef808549dd'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("CREATE SCHEMA IF NOT EXISTS triage_metadata")
    op.execute(
        "ALTER TABLE model_metadata.experiment_matrices SET SCHEMA triage_metadata;"
        + " ALTER TABLE model_metadata.experiment_models SET SCHEMA triage_metadata;"
        + " ALTER TABLE model_metadata.experiment_runs SET SCHEMA triage_metadata;"
        + " ALTER TABLE model_metadata.experiments SET SCHEMA triage_metadata;"
        + " ALTER TABLE model_metadata.list_predictions SET SCHEMA triage_metadata;"
        + " ALTER TABLE model_metadata.matrices SET SCHEMA triage_metadata;"
        + " ALTER TABLE model_metadata.model_groups SET SCHEMA triage_metadata;"
        + " ALTER TABLE model_metadata.models SET SCHEMA triage_metadata;"
        + " ALTER TABLE model_metadata.subsets SET SCHEMA triage_metadata;"
    )

    op.execute("DROP SCHEMA IF EXISTS model_metadata")

    ## We update (replace) the function
    group_proc_filename = os.path.join(
        os.path.dirname(__file__), "../../sql/model_group_stored_procedure.sql"
    )
    with open(group_proc_filename) as fd:
        stmt = fd.read()
        op.execute(stmt)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("CREATE SCHEMA IF NOT EXISTS model_metadata")

    op.execute(
        "ALTER TABLE triage_metadata.experiment_matrices SET SCHEMA model_metadata;"
        + " ALTER TABLE triage_metadata.experiment_models SET SCHEMA model_metadata;"
        + " ALTER TABLE triage_metadata.experiment_runs SET SCHEMA model_metadata;"
        + " ALTER TABLE triage_metadata.experiments SET SCHEMA model_metadata;"
        + " ALTER TABLE triage_metadata.matrices SET SCHEMA model_metadata;"
        + " ALTER TABLE triage_metadata.model_groups SET SCHEMA model_metadata;"
        + " ALTER TABLE triage_metadata.models SET SCHEMA model_metadata;"
        + " ALTER TABLE triage_metadata.subsets SET SCHEMA model_metadata;"
    )

    # ### end Alembic commands ###



import testing.postgresql
from sqlalchemy import create_engine

from triage.component.results_schema import Base


def test_full_schema():
    with testing.postgresql.Postgresql() as postgres:
        engine = create_engine(postgres.url())
        Base.metadata.create_all(bind=engine)



# coding: utf-8

import sqlalchemy
import wrapt
from contextlib import contextmanager
from sqlalchemy.orm import Session
from sqlalchemy.engine.url import make_url

import json
import functools

from psycopg2.extras import DateRange, DateTimeRange
from datetime import date, datetime


def serialize_to_database(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, date):
        return str(obj.isoformat())

    if isinstance(obj, (DateRange, DateTimeRange)):
        return f"[{obj.lower}, {obj.upper}]"

    return obj


def json_dumps(d):
    return json.dumps(d, default=serialize_to_database)



class SerializableDbEngine(wrapt.ObjectProxy):
    """A sqlalchemy engine that can be serialized across process boundaries.

    Works by saving all kwargs used to create the engine and reconstructs them later.  As a result, the state won't be saved upon serialization/deserialization.
    """

    __slots__ = ("url", "creator", "kwargs")

    def __init__(self, url, *, creator=sqlalchemy.create_engine, **kwargs):
        self.url = make_url(url)
        self.creator = creator
        self.kwargs = kwargs

        engine = creator(url, **kwargs)
        super().__init__(engine)

    def __reduce__(self):
        return (self.__reconstruct__, (self.url, self.creator, self.kwargs))

    def __reduce_ex__(self, protocol):
        # wrapt requires reduce_ex to be implemented
        return self.__reduce__()

    @classmethod
    def __reconstruct__(cls, url, creator, kwargs):
        return cls(url, creator=creator, **kwargs)


create_engine = functools.partial(SerializableDbEngine, json_serializer=json_dumps)

@contextmanager
def scoped_session(db_engine):
    """Provide a transactional scope around a series of operations."""
    session = Session(bind=db_engine)
    try:
        yield session
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()


@contextmanager
def get_for_update(db_engine, orm_class, primary_key):
    """ Gets object from the database to updated it """
    with scoped_session(db_engine) as session:
        obj = session.query(orm_class).get(primary_key)
        yield obj
        session.merge(obj)
